"""Tests for the MCP agent-facing tools."""

import json
from collections.abc import Generator
from pathlib import Path
from unittest.mock import AsyncMock, patch

import pytest

from docketeer.tools import ToolContext, registry
from docketeer_mcp.manager import MCPClientManager, MCPToolInfo


@pytest.fixture(autouse=True)
def fresh_manager() -> Generator[MCPClientManager]:
    """Replace the module-level manager with a fresh instance for each test."""
    fresh = MCPClientManager()
    with (
        patch("docketeer_mcp.tools.manager", fresh),
        patch("docketeer_mcp.prompt.manager", fresh),
    ):
        yield fresh


@pytest.fixture()
def data_dir(tmp_path: Path) -> Generator[Path]:
    d = tmp_path / "data"
    d.mkdir()
    with patch("docketeer_mcp.config.environment") as mock_env:
        mock_env.DATA_DIR = d
        yield d


@pytest.fixture()
def mcp_dir(data_dir: Path) -> Path:
    d = data_dir / "mcp"
    d.mkdir()
    return d


def _write_server(mcp_dir: Path, name: str, data: dict) -> None:
    (mcp_dir / f"{name}.json").write_text(json.dumps(data))


async def test_list_mcp_servers_none(tool_context: ToolContext, data_dir: Path):
    result = await registry.execute("list_mcp_servers", {}, tool_context)
    assert "No MCP servers configured" in result


async def test_list_mcp_servers_disconnected(tool_context: ToolContext, mcp_dir: Path):
    _write_server(mcp_dir, "time", {"command": "uvx", "args": ["mcp-server-time"]})
    result = await registry.execute("list_mcp_servers", {}, tool_context)
    assert "**time**" in result
    assert "disconnected" in result


async def test_list_mcp_servers_connected(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    _write_server(mcp_dir, "time", {"command": "uvx"})
    fresh_manager._tools["time"] = [
        MCPToolInfo(server="time", name="t1", description="", input_schema={})
    ]
    fresh_manager._clients["time"] = object()  # type: ignore[assignment]

    result = await registry.execute("list_mcp_servers", {}, tool_context)
    assert "connected (1 tools)" in result


async def test_list_mcp_servers_http(tool_context: ToolContext, mcp_dir: Path):
    _write_server(mcp_dir, "api", {"url": "https://example.com/mcp"})
    result = await registry.execute("list_mcp_servers", {}, tool_context)
    assert "https://example.com/mcp" in result


async def test_connect_already_connected(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    fresh_manager._clients["s"] = object()  # type: ignore[assignment]
    result = await registry.execute("connect_mcp_server", {"name": "s"}, tool_context)
    assert "Already connected" in result


async def test_connect_not_configured(tool_context: ToolContext, data_dir: Path):
    result = await registry.execute(
        "connect_mcp_server", {"name": "missing"}, tool_context
    )
    assert "No server configured" in result


async def test_connect_success(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    _write_server(mcp_dir, "time", {"command": "uvx"})
    tools = [
        MCPToolInfo(
            server="time", name="get_time", description="Gets the time", input_schema={}
        )
    ]
    fresh_manager.connect = AsyncMock(return_value=tools)  # type: ignore[method-assign]

    result = await registry.execute(
        "connect_mcp_server", {"name": "time"}, tool_context
    )
    assert "1 tools" in result
    assert "**get_time**" in result


async def test_connect_no_tools(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    _write_server(mcp_dir, "empty", {"command": "echo"})
    fresh_manager.connect = AsyncMock(return_value=[])  # type: ignore[method-assign]

    result = await registry.execute(
        "connect_mcp_server", {"name": "empty"}, tool_context
    )
    assert "no tools found" in result


async def test_connect_failure(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    _write_server(mcp_dir, "bad", {"command": "false"})
    fresh_manager.connect = AsyncMock(side_effect=RuntimeError("connection refused"))  # type: ignore[method-assign]

    result = await registry.execute("connect_mcp_server", {"name": "bad"}, tool_context)
    assert "Failed to connect" in result
    assert "connection refused" in result


async def test_disconnect_not_connected(tool_context: ToolContext):
    result = await registry.execute(
        "disconnect_mcp_server", {"name": "x"}, tool_context
    )
    assert "Not connected" in result


async def test_disconnect_success(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    fresh_manager._clients["s"] = object()  # type: ignore[assignment]
    fresh_manager._tools["s"] = []
    fresh_manager.disconnect = AsyncMock()  # type: ignore[method-assign]

    result = await registry.execute(
        "disconnect_mcp_server", {"name": "s"}, tool_context
    )
    assert "Disconnected" in result


async def test_search_no_results(tool_context: ToolContext):
    result = await registry.execute("search_mcp_tools", {"query": "time"}, tool_context)
    assert "No tools matching" in result


async def test_search_with_results(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    fresh_manager._tools["s"] = [
        MCPToolInfo(
            server="s",
            name="get_time",
            description="Gets the current time",
            input_schema={"type": "object", "properties": {"tz": {"type": "string"}}},
        ),
    ]
    result = await registry.execute("search_mcp_tools", {"query": "time"}, tool_context)
    assert "s / get_time" in result
    assert "Gets the current time" in result
    assert '"tz"' in result


async def test_search_no_description(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    fresh_manager._tools["s"] = [
        MCPToolInfo(server="s", name="bare_tool", description="", input_schema={}),
    ]
    result = await registry.execute("search_mcp_tools", {"query": "bare"}, tool_context)
    assert "bare_tool" in result


async def test_use_tool_success(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    fresh_manager.call_tool = AsyncMock(return_value="42")  # type: ignore[method-assign]

    result = await registry.execute(
        "use_mcp_tool",
        {"server": "s", "tool": "add", "arguments": '{"a": 1, "b": 2}'},
        tool_context,
    )
    assert result == "42"
    fresh_manager.call_tool.assert_called_once_with("s", "add", {"a": 1, "b": 2})  # type: ignore[union-attr]


async def test_use_tool_bad_json(tool_context: ToolContext):
    result = await registry.execute(
        "use_mcp_tool",
        {"server": "s", "tool": "t", "arguments": "not json"},
        tool_context,
    )
    assert "Invalid JSON" in result


async def test_use_tool_error(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    fresh_manager.call_tool = AsyncMock(side_effect=RuntimeError("timeout"))  # type: ignore[method-assign]

    result = await registry.execute(
        "use_mcp_tool",
        {"server": "s", "tool": "t"},
        tool_context,
    )
    assert "Error calling s/t" in result


async def test_add_server_stdio(tool_context: ToolContext, mcp_dir: Path):
    result = await registry.execute(
        "add_mcp_server",
        {
            "name": "time",
            "command": "uvx",
            "args": '["mcp-server-time"]',
            "env": '{"TZ": "UTC"}',
        },
        tool_context,
    )
    assert "Saved server 'time'" in result
    assert "command" in result
    data = json.loads((mcp_dir / "time.json").read_text())
    assert data["command"] == "uvx"


async def test_add_server_http(tool_context: ToolContext, mcp_dir: Path):
    result = await registry.execute(
        "add_mcp_server",
        {"name": "api", "url": "https://example.com/mcp"},
        tool_context,
    )
    assert "Saved server 'api'" in result
    assert "url" in result


async def test_add_server_no_command_or_url(tool_context: ToolContext, data_dir: Path):
    result = await registry.execute("add_mcp_server", {"name": "empty"}, tool_context)
    assert "Must provide either command" in result


async def test_add_server_bad_args_json(tool_context: ToolContext, data_dir: Path):
    result = await registry.execute(
        "add_mcp_server",
        {"name": "t", "command": "echo", "args": "not json"},
        tool_context,
    )
    assert "Invalid args JSON" in result


async def test_add_server_bad_env_json(tool_context: ToolContext, data_dir: Path):
    result = await registry.execute(
        "add_mcp_server",
        {"name": "t", "command": "echo", "env": "{bad}"},
        tool_context,
    )
    assert "Invalid env JSON" in result


async def test_add_server_bad_headers_json(tool_context: ToolContext, data_dir: Path):
    result = await registry.execute(
        "add_mcp_server",
        {"name": "t", "url": "https://x.com", "headers": "nope"},
        tool_context,
    )
    assert "Invalid headers JSON" in result


async def test_add_server_invalid_name(tool_context: ToolContext, data_dir: Path):
    result = await registry.execute(
        "add_mcp_server",
        {"name": "bad name!", "command": "echo"},
        tool_context,
    )
    assert "Invalid server name" in result


async def test_remove_server_exists(tool_context: ToolContext, mcp_dir: Path):
    (mcp_dir / "old.json").write_text("{}")
    result = await registry.execute("remove_mcp_server", {"name": "old"}, tool_context)
    assert "Removed server 'old'" in result
    assert not (mcp_dir / "old.json").exists()


async def test_remove_server_missing(tool_context: ToolContext, data_dir: Path):
    result = await registry.execute("remove_mcp_server", {"name": "gone"}, tool_context)
    assert "No server configured" in result


async def test_remove_server_disconnects_first(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    (mcp_dir / "active.json").write_text("{}")
    fresh_manager._clients["active"] = object()  # type: ignore[assignment]
    fresh_manager._tools["active"] = []
    fresh_manager.disconnect = AsyncMock()  # type: ignore[method-assign]

    result = await registry.execute(
        "remove_mcp_server", {"name": "active"}, tool_context
    )
    assert "Removed" in result
    fresh_manager.disconnect.assert_called_once_with("active")  # type: ignore[union-attr]


async def test_remove_server_cancels_oauth_refresh(
    tool_context: ToolContext, mcp_dir: Path
):
    _write_server(
        mcp_dir, "api", {"url": "https://api.example.com/mcp", "auth": "mcp/api/token"}
    )

    mock_docket = AsyncMock()
    with patch("docketeer_mcp.tools.current_docket", return_value=mock_docket):
        result = await registry.execute(
            "remove_mcp_server", {"name": "api"}, tool_context
        )

    assert "Removed" in result
    mock_docket.cancel.assert_called_once_with("mcp-refresh-mcp/api/token")


async def test_remove_server_cancel_refresh_ignores_errors(
    tool_context: ToolContext, mcp_dir: Path
):
    _write_server(
        mcp_dir, "api", {"url": "https://api.example.com/mcp", "auth": "mcp/api/token"}
    )

    mock_docket = AsyncMock()
    mock_docket.cancel = AsyncMock(side_effect=RuntimeError("no such task"))
    with patch("docketeer_mcp.tools.current_docket", return_value=mock_docket):
        result = await registry.execute(
            "remove_mcp_server", {"name": "api"}, tool_context
        )

    assert "Removed" in result



"""Edge case tests for MCP OAuth-related tools."""

import json
from collections.abc import Generator
from pathlib import Path
from unittest.mock import AsyncMock, patch

import pytest

from docketeer.tools import ToolContext, registry
from docketeer.vault import Vault
from docketeer_mcp.manager import MCPClientManager
from docketeer_mcp.oauth import PendingOAuth


@pytest.fixture(autouse=True)
def fresh_manager() -> Generator[MCPClientManager]:
    """Replace the module-level manager with a fresh instance for each test."""
    fresh = MCPClientManager()
    with (
        patch("docketeer_mcp.tools.manager", fresh),
        patch("docketeer_mcp.prompt.manager", fresh),
    ):
        yield fresh


@pytest.fixture()
def data_dir(tmp_path: Path) -> Generator[Path]:
    d = tmp_path / "data"
    d.mkdir()
    with patch("docketeer_mcp.config.environment") as mock_env:
        mock_env.DATA_DIR = d
        yield d


@pytest.fixture()
def mcp_dir(data_dir: Path) -> Path:
    d = data_dir / "mcp"
    d.mkdir()
    return d


def _write_server(mcp_dir: Path, name: str, data: dict) -> None:
    (mcp_dir / f"{name}.json").write_text(json.dumps(data))


async def test_connect_with_auth_vault_resolve_error(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When vault.resolve fails, return error message."""
    _write_server(
        mcp_dir, "api", {"url": "https://api.example.com/mcp", "auth": "mcp/api/token"}
    )

    mock_vault = AsyncMock(spec=Vault)
    mock_vault.resolve = AsyncMock(side_effect=RuntimeError("secret not found"))
    tool_context.vault = mock_vault

    result = await registry.execute("connect_mcp_server", {"name": "api"}, tool_context)
    assert "Failed to resolve" in result
    assert "secret not found" in result


async def test_connect_with_auth_connect_failure(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When vault resolves but connect() fails, return error."""
    _write_server(
        mcp_dir, "api", {"url": "https://api.example.com/mcp", "auth": "mcp/api/token"}
    )

    mock_vault = AsyncMock(spec=Vault)
    mock_vault.resolve = AsyncMock(return_value="token_123")
    tool_context.vault = mock_vault

    fresh_manager.connect = AsyncMock(side_effect=RuntimeError("connection failed"))  # type: ignore[method-assign]

    result = await registry.execute("connect_mcp_server", {"name": "api"}, tool_context)
    assert "Failed to connect" in result


async def test_connect_with_auth_no_tools(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When auth connect succeeds but no tools found."""
    _write_server(
        mcp_dir, "api", {"url": "https://api.example.com/mcp", "auth": "mcp/api/token"}
    )

    mock_vault = AsyncMock(spec=Vault)
    mock_vault.resolve = AsyncMock(return_value="token_123")
    tool_context.vault = mock_vault

    fresh_manager.connect = AsyncMock(return_value=[])  # type: ignore[method-assign]

    result = await registry.execute("connect_mcp_server", {"name": "api"}, tool_context)
    assert "no tools found" in result


async def test_connect_http_auth_discovery_failure(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When OAuth discovery fails, return error."""
    _write_server(mcp_dir, "api", {"url": "https://api.example.com/mcp"})

    with (
        patch("docketeer_mcp.tools._check_auth_required", return_value=True),
        patch(
            "docketeer_mcp.tools.discover_oauth_metadata",
            side_effect=RuntimeError("no metadata"),
        ),
    ):
        result = await registry.execute(
            "connect_mcp_server", {"name": "api"}, tool_context
        )

    assert "OAuth discovery failed" in result


async def test_connect_http_auth_no_reg_no_client_id(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When no registration endpoint and no client_id provided, return error."""
    _write_server(mcp_dir, "api", {"url": "https://api.example.com/mcp"})

    with (
        patch("docketeer_mcp.tools._check_auth_required", return_value=True),
        patch(
            "docketeer_mcp.tools.discover_oauth_metadata",
            return_value=(
                "https://api.example.com/authorize",
                "https://api.example.com/token",
                None,
                None,
            ),
        ),
    ):
        result = await registry.execute(
            "connect_mcp_server", {"name": "api"}, tool_context
        )

    assert "no registration endpoint" in result.lower()
    assert "client_id" in result.lower()


async def test_connect_http_auth_registration_failure(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When client registration fails, return error."""
    _write_server(mcp_dir, "api", {"url": "https://api.example.com/mcp"})

    with (
        patch("docketeer_mcp.tools._check_auth_required", return_value=True),
        patch(
            "docketeer_mcp.tools.discover_oauth_metadata",
            return_value=(
                "https://api.example.com/authorize",
                "https://api.example.com/token",
                "https://api.example.com/register",
                "read",
            ),
        ),
        patch(
            "docketeer_mcp.tools.register_client",
            side_effect=RuntimeError("reg failed"),
        ),
    ):
        result = await registry.execute(
            "connect_mcp_server", {"name": "api"}, tool_context
        )

    assert "registration failed" in result.lower()


async def test_mcp_oauth_complete_no_code_in_url(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    """When redirect URL has no code parameter, return error."""
    mock_vault = AsyncMock(spec=Vault)
    tool_context.vault = mock_vault

    fresh_manager._pending_oauth["s"] = PendingOAuth(
        server_url="https://example.com/mcp",
        authorization_endpoint="https://example.com/authorize",
        token_endpoint="https://example.com/token",
        code_verifier="v",
        code_challenge="c",
        state="st",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="cid",
    )

    result = await registry.execute(
        "mcp_oauth_complete",
        {
            "server": "s",
            "redirect_url": "http://127.0.0.1:3141/callback?state=st",
            "token_secret": "mcp/s/token",
        },
        tool_context,
    )
    assert "no authorization code" in result.lower()


async def test_mcp_oauth_complete_exchange_failure(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    """When exchange_code raises, return error."""
    mock_vault = AsyncMock(spec=Vault)
    tool_context.vault = mock_vault

    fresh_manager._pending_oauth["s"] = PendingOAuth(
        server_url="https://example.com/mcp",
        authorization_endpoint="https://example.com/authorize",
        token_endpoint="https://example.com/token",
        code_verifier="v",
        code_challenge="c",
        state="st",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="cid",
    )

    with patch(
        "docketeer_mcp.tools.exchange_code", side_effect=RuntimeError("invalid_grant")
    ):
        result = await registry.execute(
            "mcp_oauth_complete",
            {
                "server": "s",
                "redirect_url": "http://127.0.0.1:3141/callback?code=x&state=st",
                "token_secret": "mcp/s/token",
            },
            tool_context,
        )
    assert "token exchange failed" in result.lower()


async def test_mcp_oauth_complete_no_access_token(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    """When token response has no access_token, return error."""
    mock_vault = AsyncMock(spec=Vault)
    tool_context.vault = mock_vault

    fresh_manager._pending_oauth["s"] = PendingOAuth(
        server_url="https://example.com/mcp",
        authorization_endpoint="https://example.com/authorize",
        token_endpoint="https://example.com/token",
        code_verifier="v",
        code_challenge="c",
        state="st",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="cid",
    )

    with patch(
        "docketeer_mcp.tools.exchange_code",
        return_value={
            "token_type": "Bearer",
        },
    ):
        result = await registry.execute(
            "mcp_oauth_complete",
            {
                "server": "s",
                "redirect_url": "http://127.0.0.1:3141/callback?code=x&state=st",
                "token_secret": "mcp/s/token",
            },
            tool_context,
        )
    assert "no access_token" in result.lower()


async def test_mcp_oauth_complete_server_not_in_config(
    tool_context: ToolContext, data_dir: Path, fresh_manager: MCPClientManager
):
    """When server is not in config during complete, still succeeds."""
    mock_vault = AsyncMock(spec=Vault)
    tool_context.vault = mock_vault

    fresh_manager._pending_oauth["ghost"] = PendingOAuth(
        server_url="https://ghost.example.com/mcp",
        authorization_endpoint="https://ghost.example.com/authorize",
        token_endpoint="https://ghost.example.com/token",
        code_verifier="v",
        code_challenge="c",
        state="st",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="cid",
    )

    with patch(
        "docketeer_mcp.tools.exchange_code",
        return_value={
            "access_token": "at_123",
            "token_type": "Bearer",
        },
    ):
        result = await registry.execute(
            "mcp_oauth_complete",
            {
                "server": "ghost",
                "redirect_url": "http://127.0.0.1:3141/callback?code=x&state=st",
                "token_secret": "mcp/ghost/token",
            },
            tool_context,
        )

    assert "complete" in result.lower()
    mock_vault.store.assert_any_call("mcp/ghost/token", "at_123")
    assert "ghost" not in fresh_manager._pending_oauth


async def test_mcp_oauth_complete_schedule_refresh_failure(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When scheduling refresh fails, complete still succeeds."""
    _write_server(mcp_dir, "api", {"url": "https://api.example.com/mcp"})

    mock_vault = AsyncMock(spec=Vault)
    tool_context.vault = mock_vault

    fresh_manager._pending_oauth["api"] = PendingOAuth(
        server_url="https://api.example.com/mcp",
        authorization_endpoint="https://api.example.com/authorize",
        token_endpoint="https://api.example.com/token",
        code_verifier="v",
        code_challenge="c",
        state="st",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="cid",
        client_secret="csecret",
    )

    with (
        patch(
            "docketeer_mcp.tools.exchange_code",
            return_value={
                "access_token": "at_new",
                "token_type": "Bearer",
                "expires_in": 3600,
                "refresh_token": "rt_new",
            },
        ),
        patch(
            "docketeer_mcp.tools._schedule_token_refresh",
            side_effect=RuntimeError("no docket"),
        ),
    ):
        result = await registry.execute(
            "mcp_oauth_complete",
            {
                "server": "api",
                "redirect_url": "http://127.0.0.1:3141/callback?code=x&state=st",
                "token_secret": "mcp/api/token",
            },
            tool_context,
        )

    assert "complete" in result.lower()
    mock_vault.store.assert_any_call("mcp/api/token", "at_new")
    mock_vault.store.assert_any_call("mcp/api/token/refresh", "rt_new")



"""MCP server configuration loading and saving."""

import json
import re
from dataclasses import dataclass, field
from pathlib import Path

from docketeer import environment

_NAME_PATTERN = re.compile(r"^[a-zA-Z_][a-zA-Z0-9_-]*$")


def _mcp_dir() -> Path:
    return environment.DATA_DIR / "mcp"


@dataclass
class MCPServerConfig:
    """Configuration for a single MCP server."""

    name: str

    command: str = ""
    args: list[str] = field(default_factory=list)
    env: dict[str, str] = field(default_factory=dict)

    url: str = ""
    headers: dict[str, str] = field(default_factory=dict)

    network_access: bool = False
    auth: str = ""

    @property
    def is_stdio(self) -> bool:
        return bool(self.command)

    @property
    def is_http(self) -> bool:
        return bool(self.url)


def _validate_name(name: str) -> None:
    if not _NAME_PATTERN.match(name):
        raise ValueError(
            f"Invalid server name {name!r}: must start with a letter or underscore "
            f"and contain only letters, digits, underscores, and hyphens"
        )


def load_servers() -> dict[str, MCPServerConfig]:
    """Load all server configs from the data directory."""
    mcp_dir = _mcp_dir()
    if not mcp_dir.is_dir():
        return {}

    servers: dict[str, MCPServerConfig] = {}
    for path in sorted(mcp_dir.glob("*.json")):
        name = path.stem
        try:
            data = json.loads(path.read_text())
        except (json.JSONDecodeError, OSError):
            continue

        servers[name] = MCPServerConfig(
            name=name,
            command=data.get("command", ""),
            args=data.get("args", []),
            env=data.get("env", {}),
            url=data.get("url", ""),
            headers=data.get("headers", {}),
            network_access=data.get("networkAccess", False),
            auth=data.get("auth", ""),
        )
    return servers


def save_server(config: MCPServerConfig) -> None:
    """Write a server config to disk."""
    _validate_name(config.name)

    mcp_dir = _mcp_dir()
    mcp_dir.mkdir(parents=True, exist_ok=True)

    data: dict[str, object] = {}
    if config.command:
        data["command"] = config.command
        if config.args:
            data["args"] = config.args
        if config.env:
            data["env"] = config.env
        data["networkAccess"] = config.network_access
    elif config.url:
        data["url"] = config.url
        if config.headers:
            data["headers"] = config.headers

    if config.auth:
        data["auth"] = config.auth

    path = mcp_dir / f"{config.name}.json"
    path.write_text(json.dumps(data, indent=2) + "\n")


def remove_server(name: str) -> bool:
    """Delete a server config file. Returns True if the file existed."""
    path = _mcp_dir() / f"{name}.json"
    if not path.is_file():
        return False
    path.unlink()
    return True



"""Shared test fixtures for docketeer-mcp."""

from collections.abc import Iterator
from pathlib import Path
from unittest.mock import patch

import pytest

from docketeer import environment
from docketeer.tools import ToolContext


@pytest.fixture(autouse=True)
def _isolated_data_dir(tmp_path: Path) -> Iterator[None]:
    """Isolate tests from the real data directory."""
    with patch.object(environment, "DATA_DIR", tmp_path / "data"):
        yield


@pytest.fixture()
def workspace(tmp_path: Path) -> Path:
    ws = tmp_path / "memory"
    ws.mkdir()
    return ws


@pytest.fixture()
def tool_context(workspace: Path) -> ToolContext:
    return ToolContext(workspace=workspace, room_id="room1")



"""Tests for MCP OAuth refresh task."""

from unittest.mock import AsyncMock, patch

from docketeer_mcp.tasks import mcp_oauth_refresh


def test_mcp_task_collections_exported():
    from docketeer_mcp.tasks import mcp_task_collections

    assert "docketeer_mcp.tasks:mcp_tasks" in mcp_task_collections


async def test_mcp_oauth_refresh_success():
    """Happy path: resolve refresh token, call refresh, store new tokens."""
    mock_vault = AsyncMock()
    mock_vault.resolve = AsyncMock(return_value="old_refresh_token")

    mock_docket = AsyncMock()
    mock_docket.add = lambda *a, **kw: AsyncMock()

    with patch(
        "docketeer_mcp.tasks.refresh_access_token",
        return_value={
            "access_token": "new_at",
            "token_type": "Bearer",
            "expires_in": 7200,
        },
    ):
        await mcp_oauth_refresh(
            token_secret="mcp/api/token",
            token_endpoint="https://api.example.com/token",
            client_id="cid",
            client_secret="csecret",
            expires_in=3600,
            vault=mock_vault,
            docket=mock_docket,
        )

    mock_vault.store.assert_any_call("mcp/api/token", "new_at")


async def test_mcp_oauth_refresh_with_rotated_refresh():
    """When server returns a new refresh token, store it."""
    mock_vault = AsyncMock()
    mock_vault.resolve = AsyncMock(return_value="old_refresh_token")

    mock_docket = AsyncMock()
    mock_docket.add = lambda *a, **kw: AsyncMock()

    with patch(
        "docketeer_mcp.tasks.refresh_access_token",
        return_value={
            "access_token": "new_at",
            "token_type": "Bearer",
            "expires_in": 7200,
            "refresh_token": "new_rt",
        },
    ):
        await mcp_oauth_refresh(
            token_secret="mcp/api/token",
            token_endpoint="https://api.example.com/token",
            client_id="cid",
            vault=mock_vault,
            docket=mock_docket,
        )

    mock_vault.store.assert_any_call("mcp/api/token", "new_at")
    mock_vault.store.assert_any_call("mcp/api/token/refresh", "new_rt")


async def test_mcp_oauth_refresh_no_access_token():
    """When response has no access_token, skip storing it."""
    mock_vault = AsyncMock()
    mock_vault.resolve = AsyncMock(return_value="old_refresh_token")

    mock_docket = AsyncMock()
    mock_docket.add = lambda *a, **kw: AsyncMock()

    with patch(
        "docketeer_mcp.tasks.refresh_access_token",
        return_value={
            "token_type": "Bearer",
            "expires_in": 7200,
        },
    ):
        await mcp_oauth_refresh(
            token_secret="mcp/api/token",
            token_endpoint="https://api.example.com/token",
            client_id="cid",
            vault=mock_vault,
            docket=mock_docket,
        )

    # Should not store an empty access token
    for call in mock_vault.store.call_args_list:
        assert call.args[0] != "mcp/api/token" or call.args[1] != ""


async def test_mcp_oauth_refresh_resolve_failure():
    """When refresh token can't be resolved, log and return without crashing."""
    mock_vault = AsyncMock()
    mock_vault.resolve = AsyncMock(side_effect=RuntimeError("vault error"))

    mock_docket = AsyncMock()

    # Should not raise
    await mcp_oauth_refresh(
        token_secret="mcp/api/token",
        token_endpoint="https://api.example.com/token",
        client_id="cid",
        vault=mock_vault,
        docket=mock_docket,
    )

    mock_vault.store.assert_not_called()


async def test_mcp_oauth_refresh_api_failure():
    """When the refresh API call fails, log and return without crashing."""
    mock_vault = AsyncMock()
    mock_vault.resolve = AsyncMock(return_value="old_refresh_token")

    mock_docket = AsyncMock()

    with patch(
        "docketeer_mcp.tasks.refresh_access_token",
        side_effect=RuntimeError("refresh failed"),
    ):
        await mcp_oauth_refresh(
            token_secret="mcp/api/token",
            token_endpoint="https://api.example.com/token",
            client_id="cid",
            vault=mock_vault,
            docket=mock_docket,
        )

    mock_vault.store.assert_not_called()



"""Custom MCP transport using the CommandExecutor abstraction."""

import contextlib
import logging
from collections.abc import AsyncIterator
from typing import Any

import anyio
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from fastmcp.client.transports import ClientTransport
from mcp import ClientSession
from mcp.shared.message import SessionMessage
from mcp.types import JSONRPCMessage

from docketeer.executor import CommandExecutor, Mount, RunningProcess

log = logging.getLogger(__name__)


class ExecutorTransport(ClientTransport):
    """Launches an MCP server through a CommandExecutor and speaks stdio JSONRPC."""

    def __init__(
        self,
        executor: CommandExecutor,
        command: list[str],
        env: dict[str, str] | None = None,
        mounts: list[Mount] | None = None,
        network_access: bool = False,
    ) -> None:
        self._executor = executor
        self._command = command
        self._env = env
        self._mounts = mounts
        self._network_access = network_access

    @contextlib.asynccontextmanager
    async def connect_session(
        self, **session_kwargs: Any
    ) -> AsyncIterator[ClientSession]:
        label = self._command[0] if self._command else "unknown"
        log.info("Starting MCP server: %s", " ".join(self._command))

        process = await self._executor.start(
            self._command,
            env=self._env or None,
            mounts=self._mounts or None,
            network_access=self._network_access,
        )
        try:
            read_send, read_recv = anyio.create_memory_object_stream[
                SessionMessage | Exception
            ](0)
            write_send, write_recv = anyio.create_memory_object_stream[SessionMessage](
                0
            )

            async with anyio.create_task_group() as tg:
                tg.start_soon(_stdout_reader, process, read_send)
                tg.start_soon(_stdin_writer, process, write_recv)
                tg.start_soon(_stderr_logger, process, label)

                try:
                    async with read_recv, write_send:
                        async with ClientSession(
                            read_recv, write_send, **session_kwargs
                        ) as session:
                            yield session
                finally:
                    tg.cancel_scope.cancel()
        finally:
            log.info("Terminating MCP server: %s", " ".join(self._command))
            process.terminate()


async def _stdout_reader(
    process: RunningProcess,
    send_stream: MemoryObjectSendStream[SessionMessage | Exception],
) -> None:
    """Read newline-delimited JSONRPC from the process stdout into the memory stream."""
    assert process.stdout is not None
    async with send_stream:
        buffer = b""
        while True:
            chunk = await process.stdout.read(65_536)
            if not chunk:
                break
            lines = (buffer + chunk).split(b"\n")
            buffer = lines.pop()
            for line in lines:
                line = line.strip()
                if not line:
                    continue
                try:
                    message = JSONRPCMessage.model_validate_json(line)
                except Exception as exc:
                    await send_stream.send(exc)
                    continue
                await send_stream.send(SessionMessage(message))


async def _stdin_writer(
    process: RunningProcess,
    recv_stream: MemoryObjectReceiveStream[SessionMessage],
) -> None:
    """Read SessionMessages from the memory stream and write as JSONRPC to process stdin."""
    assert process.stdin is not None
    async with recv_stream:
        async for session_message in recv_stream:
            data = session_message.message.model_dump_json(
                by_alias=True, exclude_none=True
            )
            process.stdin.write((data + "\n").encode())
            await process.stdin.drain()


async def _stderr_logger(
    process: RunningProcess,
    label: str,
) -> None:
    """Read stderr from the MCP server process and log it."""
    if not process.stderr:
        return
    while True:
        chunk = await process.stderr.read(65_536)
        if not chunk:
            break
        for line in chunk.decode(errors="replace").splitlines():
            line = line.strip()
            if line:
                log.info("[%s] %s", label, line)



"""Tests for MCP server configuration."""

import json
from collections.abc import Generator
from pathlib import Path
from unittest.mock import patch

import pytest

from docketeer_mcp.config import (
    MCPServerConfig,
    _validate_name,
    load_servers,
    remove_server,
    save_server,
)


@pytest.fixture()
def data_dir(tmp_path: Path) -> Generator[Path]:
    d = tmp_path / "data"
    d.mkdir()
    with patch("docketeer_mcp.config.environment") as mock_env:
        mock_env.DATA_DIR = d
        yield d


@pytest.fixture()
def mcp_dir(data_dir: Path) -> Path:
    d = data_dir / "mcp"
    d.mkdir()
    return d


def test_config_is_stdio():
    c = MCPServerConfig(name="t", command="uvx", args=["server"])
    assert c.is_stdio
    assert not c.is_http


def test_config_is_http():
    c = MCPServerConfig(name="t", url="https://example.com/mcp")
    assert c.is_http
    assert not c.is_stdio


def test_config_neither():
    c = MCPServerConfig(name="t")
    assert not c.is_stdio
    assert not c.is_http


def test_validate_name_valid():
    for name in ["time", "my_server", "a-b-c", "_private", "A1"]:
        _validate_name(name)


def test_validate_name_invalid():
    for name in ["", "123", "has space", "no!bang", "a/b"]:
        with pytest.raises(ValueError, match="Invalid server name"):
            _validate_name(name)


def test_load_servers_no_dir(data_dir: Path):
    assert load_servers() == {}


def test_load_servers_stdio(mcp_dir: Path):
    (mcp_dir / "time.json").write_text(
        json.dumps(
            {
                "command": "uvx",
                "args": ["mcp-server-time"],
                "env": {"TZ": "UTC"},
                "networkAccess": True,
            }
        )
    )
    servers = load_servers()
    assert "time" in servers
    s = servers["time"]
    assert s.name == "time"
    assert s.command == "uvx"
    assert s.args == ["mcp-server-time"]
    assert s.env == {"TZ": "UTC"}
    assert s.network_access is True
    assert s.is_stdio


def test_load_servers_http(mcp_dir: Path):
    (mcp_dir / "weather.json").write_text(
        json.dumps(
            {
                "url": "https://weather.example.com/mcp",
                "headers": {"Authorization": "Bearer tok"},
            }
        )
    )
    servers = load_servers()
    assert "weather" in servers
    s = servers["weather"]
    assert s.url == "https://weather.example.com/mcp"
    assert s.headers == {"Authorization": "Bearer tok"}
    assert s.is_http


def test_load_servers_skips_bad_json(mcp_dir: Path):
    (mcp_dir / "good.json").write_text('{"command": "echo"}')
    (mcp_dir / "bad.json").write_text("not json{{{")
    servers = load_servers()
    assert "good" in servers
    assert "bad" not in servers


def test_load_servers_sorted(mcp_dir: Path):
    (mcp_dir / "beta.json").write_text('{"command": "b"}')
    (mcp_dir / "alpha.json").write_text('{"command": "a"}')
    servers = load_servers()
    assert list(servers.keys()) == ["alpha", "beta"]


def test_save_server_stdio(mcp_dir: Path):
    config = MCPServerConfig(
        name="time",
        command="uvx",
        args=["mcp-server-time"],
        env={"TZ": "UTC"},
        network_access=True,
    )
    save_server(config)
    data = json.loads((mcp_dir / "time.json").read_text())
    assert data["command"] == "uvx"
    assert data["args"] == ["mcp-server-time"]
    assert data["env"] == {"TZ": "UTC"}
    assert data["networkAccess"] is True


def test_save_server_stdio_minimal(mcp_dir: Path):
    config = MCPServerConfig(name="simple", command="echo")
    save_server(config)
    data = json.loads((mcp_dir / "simple.json").read_text())
    assert data == {"command": "echo", "networkAccess": False}


def test_save_server_http(mcp_dir: Path):
    config = MCPServerConfig(
        name="api",
        url="https://api.example.com/mcp",
        headers={"X-Key": "secret"},
    )
    save_server(config)
    data = json.loads((mcp_dir / "api.json").read_text())
    assert data["url"] == "https://api.example.com/mcp"
    assert data["headers"] == {"X-Key": "secret"}


def test_save_server_http_minimal(mcp_dir: Path):
    config = MCPServerConfig(name="bare", url="https://example.com/mcp")
    save_server(config)
    data = json.loads((mcp_dir / "bare.json").read_text())
    assert data == {"url": "https://example.com/mcp"}


def test_save_server_creates_dir(data_dir: Path):
    save_server(MCPServerConfig(name="test", command="echo"))
    assert (data_dir / "mcp" / "test.json").is_file()


def test_save_server_empty_config(mcp_dir: Path):
    config = MCPServerConfig(name="empty")
    save_server(config)
    data = json.loads((mcp_dir / "empty.json").read_text())
    assert data == {}


def test_save_server_invalid_name(data_dir: Path):
    with pytest.raises(ValueError, match="Invalid server name"):
        save_server(MCPServerConfig(name="bad name!", command="echo"))


def test_remove_server_exists(mcp_dir: Path):
    (mcp_dir / "doomed.json").write_text("{}")
    assert remove_server("doomed") is True
    assert not (mcp_dir / "doomed.json").exists()


def test_remove_server_missing(mcp_dir: Path):
    assert remove_server("nonexistent") is False


# --- auth field ---


def test_config_auth_default():
    c = MCPServerConfig(name="t", url="https://example.com/mcp")
    assert c.auth == ""


def test_load_servers_with_auth(mcp_dir: Path):
    (mcp_dir / "api.json").write_text(
        json.dumps(
            {
                "url": "https://api.example.com/mcp",
                "auth": "mcp/api/token",
            }
        )
    )
    servers = load_servers()
    assert servers["api"].auth == "mcp/api/token"


def test_load_servers_without_auth(mcp_dir: Path):
    (mcp_dir / "api.json").write_text(
        json.dumps({"url": "https://api.example.com/mcp"})
    )
    servers = load_servers()
    assert servers["api"].auth == ""


def test_save_server_http_with_auth(mcp_dir: Path):
    config = MCPServerConfig(
        name="api",
        url="https://api.example.com/mcp",
        auth="mcp/api/token",
    )
    save_server(config)
    data = json.loads((mcp_dir / "api.json").read_text())
    assert data["auth"] == "mcp/api/token"


def test_save_server_http_without_auth(mcp_dir: Path):
    config = MCPServerConfig(name="api", url="https://api.example.com/mcp")
    save_server(config)
    data = json.loads((mcp_dir / "api.json").read_text())
    assert "auth" not in data



"""Tests for the ExecutorTransport."""

import json
import logging
from pathlib import Path

import anyio
import pytest
from mcp.shared.message import SessionMessage
from mcp.types import JSONRPCMessage

from docketeer.executor import CommandExecutor, Mount, RunningProcess
from docketeer_mcp.transport import (
    ExecutorTransport,
    _stderr_logger,
    _stdin_writer,
    _stdout_reader,
)


class FakeStdin:
    """Minimal async writer that collects written bytes."""

    def __init__(self) -> None:
        self._buf = bytearray()

    def write(self, data: bytes) -> None:
        self._buf.extend(data)

    async def drain(self) -> None:
        pass

    @property
    def written(self) -> bytes:
        return bytes(self._buf)


class FakeStdout:
    """Minimal async reader backed by a pre-loaded buffer."""

    def __init__(self, data: bytes) -> None:
        self._data = data
        self._pos = 0

    async def read(self, n: int) -> bytes:
        chunk = self._data[self._pos : self._pos + n]
        self._pos += n
        return chunk


class FakeProcess:
    """A minimal stand-in for RunningProcess that avoids real asyncio streams."""

    def __init__(self, stdout_data: bytes = b"", stderr_data: bytes = b"") -> None:
        self.stdin = FakeStdin()
        self.stdout = FakeStdout(stdout_data)
        self.stderr = FakeStdout(stderr_data) if stderr_data else None
        self.returncode: int | None = None
        self.terminated = False

    def terminate(self) -> None:
        self.terminated = True
        self.returncode = -15


class FakeExecutor(CommandExecutor):
    """A fake executor that returns a FakeProcess."""

    def __init__(self, stdout_data: bytes = b"") -> None:
        self._stdout_data = stdout_data
        self.last_process: FakeProcess | None = None
        self.last_command: list[str] = []
        self.last_env: dict[str, str] | None = None
        self.last_mounts: list[Mount] | None = None
        self.last_network_access: bool = False

    async def start(
        self,
        command: list[str],
        *,
        env: dict[str, str] | None = None,
        mounts: list[Mount] | None = None,
        network_access: bool = False,
        username: str | None = None,
    ) -> RunningProcess:
        self.last_command = command
        self.last_env = env
        self.last_mounts = mounts
        self.last_network_access = network_access
        self.last_process = FakeProcess(self._stdout_data)
        return self.last_process  # type: ignore[return-value]


async def test_stdout_reader_parses_jsonrpc():
    msg = {"jsonrpc": "2.0", "id": 1, "result": {"tools": []}}
    process = FakeProcess(json.dumps(msg).encode() + b"\n")

    send, recv = anyio.create_memory_object_stream[SessionMessage | Exception](10)
    await _stdout_reader(process, send)  # type: ignore[arg-type]

    async with recv:
        item = recv.receive_nowait()
        assert isinstance(item, SessionMessage)
        assert item.message.root.id == 1  # type: ignore[union-attr]


async def test_stdout_reader_sends_parse_errors():
    process = FakeProcess(b"not valid json\n")

    send, recv = anyio.create_memory_object_stream[SessionMessage | Exception](10)
    await _stdout_reader(process, send)  # type: ignore[arg-type]

    async with recv:
        item = recv.receive_nowait()
        assert isinstance(item, Exception)


async def test_stdout_reader_skips_blank_lines():
    msg = {"jsonrpc": "2.0", "id": 1, "result": {}}
    process = FakeProcess(b"\n\n" + json.dumps(msg).encode() + b"\n\n")

    send, recv = anyio.create_memory_object_stream[SessionMessage | Exception](10)
    await _stdout_reader(process, send)  # type: ignore[arg-type]

    async with recv:
        item = recv.receive_nowait()
        assert isinstance(item, SessionMessage)


async def test_stdin_writer_serializes_messages():
    process = FakeProcess()
    msg = JSONRPCMessage.model_validate({"jsonrpc": "2.0", "id": 1, "method": "ping"})
    session_msg = SessionMessage(msg)

    send, recv = anyio.create_memory_object_stream[SessionMessage](10)
    async with send:
        await send.send(session_msg)

    await _stdin_writer(process, recv)  # type: ignore[arg-type]

    parsed = json.loads(process.stdin.written.strip())
    assert parsed["jsonrpc"] == "2.0"
    assert parsed["method"] == "ping"


async def test_executor_transport_stores_params():
    executor = FakeExecutor()
    mounts = [Mount(source=Path("/src"), target=Path("/dst"))]
    transport = ExecutorTransport(
        executor=executor,
        command=["uvx", "mcp-server-test"],
        env={"KEY": "val"},
        mounts=mounts,
        network_access=True,
    )

    assert transport._executor is executor
    assert transport._command == ["uvx", "mcp-server-test"]
    assert transport._env == {"KEY": "val"}
    assert transport._mounts == mounts
    assert transport._network_access is True


async def test_executor_transport_connect_session():
    executor = FakeExecutor()
    transport = ExecutorTransport(executor=executor, command=["echo"])

    with pytest.raises(ExceptionGroup):
        async with transport.connect_session():
            raise RuntimeError("bail")

    assert executor.last_command == ["echo"]
    assert executor.last_env is None
    assert executor.last_mounts is None
    assert executor.last_network_access is False
    assert executor.last_process is not None
    assert executor.last_process.terminated


async def test_executor_transport_passes_through_env():
    executor = FakeExecutor()
    transport = ExecutorTransport(
        executor=executor, command=["echo"], env={"HOME": "/custom", "FOO": "bar"}
    )

    with pytest.raises(ExceptionGroup):
        async with transport.connect_session():
            raise RuntimeError("bail")

    assert executor.last_env == {"HOME": "/custom", "FOO": "bar"}


async def test_stderr_logger_logs_output(caplog: pytest.LogCaptureFixture):
    process = FakeProcess(stderr_data=b"Starting server...\nReady\n")

    with caplog.at_level(logging.INFO):
        await _stderr_logger(process, "test-server")  # type: ignore[arg-type]

    assert "[test-server] Starting server..." in caplog.text
    assert "[test-server] Ready" in caplog.text


async def test_stderr_logger_skips_blank_lines(caplog: pytest.LogCaptureFixture):
    process = FakeProcess(stderr_data=b"\n  \nOutput\n\n")

    with caplog.at_level(logging.INFO):
        await _stderr_logger(process, "srv")  # type: ignore[arg-type]

    assert "[srv] Output" in caplog.text
    assert caplog.text.count("[srv]") == 1


async def test_stderr_logger_handles_none_stderr():
    process = FakeProcess()  # stderr is None
    await _stderr_logger(process, "srv")  # type: ignore[arg-type]



"""System prompt provider for MCP server catalog."""

from pathlib import Path

from docketeer.prompt import SystemBlock

from .config import load_servers
from .manager import manager


def provide_mcp_catalog(workspace: Path) -> list[SystemBlock]:
    """Build an MCP server catalog block for the system prompt."""
    servers = load_servers()
    if not servers:
        return []

    connected = manager.connected_servers()

    lines = [
        "## MCP Servers",
        "",
        "Use `connect_mcp_server` to connect, `search_mcp_tools` to find tools, "
        "and `use_mcp_tool` to call them.",
        "",
        "Configured servers:",
    ]
    for name, cfg in servers.items():
        kind = f"`{cfg.command}`" if cfg.is_stdio else f"`{cfg.url}`"
        lines.append(f"- **{name}**: {kind}")

    if connected:
        parts = [f"{name} ({count} tools)" for name, count in connected.items()]
        lines.append("")
        lines.append(f"Connected: {', '.join(parts)}")

    return [SystemBlock(text="\n".join(lines))]



"""OAuth helpers for MCP server authentication.

Standalone functions that handle the OAuth discovery, registration, PKCE,
and token exchange steps. The agent orchestrates these via tools — no
subclassing of FastMCP's OAuthClientProvider needed.
"""

import hashlib
import secrets
import string
from base64 import urlsafe_b64encode
from dataclasses import dataclass
from urllib.parse import urlencode, urlparse

import httpx
from mcp.shared.auth import OAuthMetadata, ProtectedResourceMetadata

REDIRECT_URI = "http://127.0.0.1:3141/callback"


@dataclass
class PendingOAuth:
    """State held between the start and completion of an OAuth flow."""

    server_url: str
    authorization_endpoint: str
    token_endpoint: str
    code_verifier: str
    code_challenge: str
    state: str
    redirect_uri: str
    client_id: str
    client_secret: str = ""
    scopes: str = ""


def _generate_pkce() -> tuple[str, str]:
    """Generate a PKCE code_verifier and code_challenge (S256)."""
    verifier = "".join(
        secrets.choice(string.ascii_letters + string.digits + "-._~")
        for _ in range(128)
    )
    digest = hashlib.sha256(verifier.encode()).digest()
    challenge = urlsafe_b64encode(digest).decode().rstrip("=")
    return verifier, challenge


async def discover_oauth_metadata(
    server_url: str,
) -> tuple[str, str, str | None, str | None]:
    """Discover authorization and token endpoints for an MCP server.

    Tries RFC 9728 Protected Resource Metadata first, then falls back to
    legacy OASM discovery (/.well-known/oauth-authorization-server).

    Returns (authorization_endpoint, token_endpoint, registration_endpoint, scopes).
    """
    parsed = urlparse(server_url)
    base_url = f"{parsed.scheme}://{parsed.netloc}"

    async with httpx.AsyncClient() as client:
        # Try Protected Resource Metadata (path-aware, then root)
        prm_urls = []
        if parsed.path and parsed.path != "/":
            prm_urls.append(
                f"{base_url}/.well-known/oauth-protected-resource{parsed.path}"
            )
        prm_urls.append(f"{base_url}/.well-known/oauth-protected-resource")

        auth_server_url: str | None = None
        scopes: str | None = None

        for url in prm_urls:
            resp = await client.get(url)
            if resp.status_code == 200:
                prm = ProtectedResourceMetadata.model_validate_json(resp.content)
                auth_server_url = str(prm.authorization_servers[0])
                if prm.scopes_supported:
                    scopes = " ".join(prm.scopes_supported)
                break

        # Discover OAuth Authorization Server Metadata
        oasm_urls = []
        if auth_server_url:
            oasm_parsed = urlparse(auth_server_url)
            oasm_base = f"{oasm_parsed.scheme}://{oasm_parsed.netloc}"
            if oasm_parsed.path and oasm_parsed.path != "/":
                oasm_urls.append(
                    f"{oasm_base}/.well-known/oauth-authorization-server{oasm_parsed.path}"
                )
            oasm_urls.append(f"{oasm_base}/.well-known/oauth-authorization-server")
        else:
            oasm_urls.append(f"{base_url}/.well-known/oauth-authorization-server")

        for url in oasm_urls:
            resp = await client.get(url)
            if resp.status_code == 200:
                oasm = OAuthMetadata.model_validate_json(resp.content)
                reg_ep = (
                    str(oasm.registration_endpoint)
                    if oasm.registration_endpoint
                    else None
                )
                if not scopes and oasm.scopes_supported:
                    scopes = " ".join(oasm.scopes_supported)
                return (
                    str(oasm.authorization_endpoint),
                    str(oasm.token_endpoint),
                    reg_ep,
                    scopes,
                )

    raise RuntimeError(f"Could not discover OAuth metadata for {server_url}")


async def register_client(
    registration_endpoint: str,
    redirect_uri: str,
    client_name: str,
    scopes: str = "",
) -> tuple[str, str]:
    """Dynamic client registration (RFC 7591).

    Returns (client_id, client_secret).
    """
    body: dict[str, object] = {
        "client_name": client_name,
        "redirect_uris": [redirect_uri],
        "grant_types": ["authorization_code", "refresh_token"],
        "response_types": ["code"],
        "token_endpoint_auth_method": "none",
    }
    if scopes:
        body["scope"] = scopes

    async with httpx.AsyncClient() as client:
        resp = await client.post(registration_endpoint, json=body)

    if resp.status_code not in (200, 201):
        raise RuntimeError(
            f"Client registration failed ({resp.status_code}): {resp.text}"
        )

    data = resp.json()
    return data["client_id"], data.get("client_secret", "")


def build_authorization_url(pending: PendingOAuth) -> str:
    """Build the full authorization URL with PKCE and state."""
    params: dict[str, str] = {
        "response_type": "code",
        "client_id": pending.client_id,
        "redirect_uri": pending.redirect_uri,
        "state": pending.state,
        "code_challenge": pending.code_challenge,
        "code_challenge_method": "S256",
    }
    if pending.scopes:
        params["scope"] = pending.scopes

    return f"{pending.authorization_endpoint}?{urlencode(params)}"


async def exchange_code(pending: PendingOAuth, code: str) -> dict[str, object]:
    """Exchange an authorization code for tokens.

    Returns the full token response dict (access_token, token_type, etc.).
    """
    data: dict[str, str] = {
        "grant_type": "authorization_code",
        "code": code,
        "redirect_uri": pending.redirect_uri,
        "client_id": pending.client_id,
        "code_verifier": pending.code_verifier,
    }
    if pending.client_secret:
        data["client_secret"] = pending.client_secret

    async with httpx.AsyncClient() as client:
        resp = await client.post(
            pending.token_endpoint,
            data=data,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )

    if resp.status_code != 200:
        raise RuntimeError(f"Token exchange failed ({resp.status_code}): {resp.text}")

    return resp.json()


async def refresh_access_token(
    token_endpoint: str,
    refresh_token: str,
    client_id: str,
    client_secret: str = "",
) -> dict[str, object]:
    """Refresh an access token. Returns the new token response dict."""
    data: dict[str, str] = {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
        "client_id": client_id,
    }
    if client_secret:
        data["client_secret"] = client_secret

    async with httpx.AsyncClient() as client:
        resp = await client.post(
            token_endpoint,
            data=data,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )

    if resp.status_code != 200:
        raise RuntimeError(f"Token refresh failed ({resp.status_code}): {resp.text}")

    return resp.json()



"""Tests for the MCPClientManager."""

import logging
from dataclasses import dataclass
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from docketeer.executor import CommandExecutor, Mount, RunningProcess
from docketeer_mcp.config import MCPServerConfig
from docketeer_mcp.manager import (
    MCPClientManager,
    MCPToolInfo,
    _build_transport,
)
from docketeer_mcp.transport import ExecutorTransport


class FakeExecutor(CommandExecutor):
    async def start(
        self,
        command: list[str],
        *,
        env: dict[str, str] | None = None,
        mounts: list[Mount] | None = None,
        network_access: bool = False,
        username: str | None = None,
    ) -> RunningProcess:
        raise NotImplementedError


@dataclass
class FakeTool:
    name: str
    description: str | None
    inputSchema: dict


@dataclass
class FakeTextContent:
    text: str


@dataclass
class FakeImageContent:
    data: str
    mimeType: str


@dataclass
class FakeCallResult:
    content: list


def _mock_client(
    tools: list[FakeTool] | None = None, call_result: FakeCallResult | None = None
) -> MagicMock:
    client = MagicMock()
    client.__aenter__ = AsyncMock(return_value=client)
    client.__aexit__ = AsyncMock(return_value=None)
    client.list_tools = AsyncMock(return_value=tools or [])
    client.call_tool = AsyncMock(return_value=call_result or FakeCallResult(content=[]))
    return client


# --- _build_transport tests ---


def test_build_transport_stdio_with_executor():
    config = MCPServerConfig(
        name="t", command="uvx", args=["server"], env={"K": "V"}, network_access=True
    )
    executor = FakeExecutor()
    workspace = Path("/ws")
    transport = _build_transport(config, executor, workspace)
    assert isinstance(transport, ExecutorTransport)
    assert transport._command == ["uvx", "server"]
    assert transport._env == {"K": "V"}
    assert transport._network_access is True
    assert transport._mounts == [Mount(source=Path("/ws"), target=Path("/workspace"))]


def test_build_transport_stdio_with_executor_no_workspace():
    config = MCPServerConfig(name="t", command="echo")
    executor = FakeExecutor()
    transport = _build_transport(config, executor)
    assert isinstance(transport, ExecutorTransport)
    assert transport._mounts is None


def test_build_transport_stdio_with_executor_empty_env():
    config = MCPServerConfig(name="t", command="echo")
    executor = FakeExecutor()
    transport = _build_transport(config, executor)
    assert isinstance(transport, ExecutorTransport)
    assert transport._env is None


def test_build_transport_stdio_without_executor(caplog: pytest.LogCaptureFixture):
    config = MCPServerConfig(name="test", command="python", args=["-m", "server"])
    with caplog.at_level(logging.WARNING):
        transport = _build_transport(config)
    assert "without sandbox" in caplog.text
    from fastmcp.client.transports.stdio import StdioTransport

    assert isinstance(transport, StdioTransport)


def test_build_transport_stdio_without_executor_empty_env():
    config = MCPServerConfig(name="test", command="python")
    transport = _build_transport(config)
    from fastmcp.client.transports.stdio import StdioTransport

    assert isinstance(transport, StdioTransport)


def test_build_transport_http():
    config = MCPServerConfig(
        name="api",
        url="https://example.com/mcp",
        headers={"Authorization": "Bearer tok"},
    )
    transport = _build_transport(config)
    from fastmcp.client.transports import StreamableHttpTransport

    assert isinstance(transport, StreamableHttpTransport)


def test_build_transport_http_no_headers():
    config = MCPServerConfig(name="api", url="https://example.com/mcp")
    transport = _build_transport(config)
    from fastmcp.client.transports import StreamableHttpTransport

    assert isinstance(transport, StreamableHttpTransport)


def test_build_transport_invalid():
    config = MCPServerConfig(name="bad")
    with pytest.raises(ValueError, match="neither command nor url"):
        _build_transport(config)


# --- MCPClientManager tests ---


@pytest.fixture()
def mgr() -> MCPClientManager:
    return MCPClientManager()


async def test_connect_discovers_tools(mgr: MCPClientManager):
    tools = [
        FakeTool(
            name="get_time",
            description="Get current time",
            inputSchema={"type": "object"},
        ),
        FakeTool(name="convert", description=None, inputSchema={}),
    ]
    client = _mock_client(tools)
    config = MCPServerConfig(name="time", command="echo")

    with patch("docketeer_mcp.manager.Client", return_value=client):
        result = await mgr.connect("time", config)

    assert len(result) == 2
    assert result[0].name == "get_time"
    assert result[0].description == "Get current time"
    assert result[1].description == ""
    assert mgr.is_connected("time")


async def test_connect_already_connected(mgr: MCPClientManager):
    client = _mock_client()
    config = MCPServerConfig(name="s", command="echo")

    with patch("docketeer_mcp.manager.Client", return_value=client):
        await mgr.connect("s", config)

    with pytest.raises(ValueError, match="Already connected"):
        await mgr.connect("s", config)


async def test_connect_list_tools_failure_cleans_up(mgr: MCPClientManager):
    client = _mock_client()
    client.list_tools = AsyncMock(side_effect=RuntimeError("boom"))
    config = MCPServerConfig(name="bad", command="echo")

    with patch("docketeer_mcp.manager.Client", return_value=client):
        with pytest.raises(RuntimeError, match="boom"):
            await mgr.connect("bad", config)

    assert not mgr.is_connected("bad")
    client.__aexit__.assert_called_once()


async def test_disconnect(mgr: MCPClientManager):
    client = _mock_client()
    config = MCPServerConfig(name="s", command="echo")

    with patch("docketeer_mcp.manager.Client", return_value=client):
        await mgr.connect("s", config)

    await mgr.disconnect("s")
    assert not mgr.is_connected("s")
    assert client.__aexit__.call_count == 1


async def test_disconnect_not_connected(mgr: MCPClientManager):
    await mgr.disconnect("nonexistent")


async def test_disconnect_all(mgr: MCPClientManager):
    clients = {}
    for name in ("a", "b"):
        client = _mock_client()
        clients[name] = client
        config = MCPServerConfig(name=name, command="echo")
        with patch("docketeer_mcp.manager.Client", return_value=client):
            await mgr.connect(name, config)

    await mgr.disconnect_all()
    assert not mgr.is_connected("a")
    assert not mgr.is_connected("b")


async def test_connected_servers(mgr: MCPClientManager):
    tools_a = [FakeTool("t1", "desc", {}), FakeTool("t2", "desc", {})]
    tools_b = [FakeTool("t3", "desc", {})]

    for name, tools in [("a", tools_a), ("b", tools_b)]:
        client = _mock_client(tools)
        config = MCPServerConfig(name=name, command="echo")
        with patch("docketeer_mcp.manager.Client", return_value=client):
            await mgr.connect(name, config)

    servers = mgr.connected_servers()
    assert servers == {"a": 2, "b": 1}


async def test_search_tools_by_name(mgr: MCPClientManager):
    mgr._tools["s"] = [
        MCPToolInfo(
            server="s", name="get_time", description="Gets the time", input_schema={}
        ),
        MCPToolInfo(
            server="s", name="set_alarm", description="Sets an alarm", input_schema={}
        ),
    ]
    results = mgr.search_tools("time")
    assert len(results) == 1
    assert results[0].name == "get_time"


async def test_search_tools_by_description(mgr: MCPClientManager):
    mgr._tools["s"] = [
        MCPToolInfo(
            server="s", name="foo", description="Fetches weather data", input_schema={}
        ),
    ]
    results = mgr.search_tools("weather")
    assert len(results) == 1


async def test_search_tools_case_insensitive(mgr: MCPClientManager):
    mgr._tools["s"] = [
        MCPToolInfo(server="s", name="GetTime", description="", input_schema={}),
    ]
    results = mgr.search_tools("gettime")
    assert len(results) == 1


async def test_search_tools_filter_by_server(mgr: MCPClientManager):
    mgr._tools["a"] = [
        MCPToolInfo(server="a", name="tool1", description="", input_schema={})
    ]
    mgr._tools["b"] = [
        MCPToolInfo(server="b", name="tool2", description="", input_schema={})
    ]
    results = mgr.search_tools("tool", server="a")
    assert len(results) == 1
    assert results[0].server == "a"


async def test_search_tools_no_match(mgr: MCPClientManager):
    mgr._tools["s"] = [
        MCPToolInfo(server="s", name="x", description="y", input_schema={})
    ]
    assert mgr.search_tools("zzz") == []


async def test_call_tool_text_result(mgr: MCPClientManager):
    result = FakeCallResult(content=[FakeTextContent(text="hello")])
    client = _mock_client(call_result=result)
    mgr._clients["s"] = client

    output = await mgr.call_tool("s", "greet", {"name": "Chris"})
    assert output == "hello"
    client.call_tool.assert_called_once_with("greet", {"name": "Chris"})


async def test_call_tool_non_text_result(mgr: MCPClientManager):
    result = FakeCallResult(
        content=[FakeImageContent(data="abc", mimeType="image/png")]
    )
    client = _mock_client(call_result=result)
    mgr._clients["s"] = client

    output = await mgr.call_tool("s", "screenshot", {})
    assert "FakeImageContent" in output


async def test_call_tool_not_connected(mgr: MCPClientManager):
    with pytest.raises(ValueError, match="Not connected"):
        await mgr.call_tool("missing", "tool", {})



"""Tests for MCP OAuth-related tools (connect with auth, mcp_oauth_complete)."""

import json
from collections.abc import Generator
from pathlib import Path
from unittest.mock import AsyncMock, patch

import httpx
import pytest
from docket import Docket

from docketeer.tools import ToolContext, registry
from docketeer.vault import Vault
from docketeer_mcp.manager import MCPClientManager, MCPToolInfo
from docketeer_mcp.oauth import PendingOAuth


@pytest.fixture(autouse=True)
def fresh_manager() -> Generator[MCPClientManager]:
    """Replace the module-level manager with a fresh instance for each test."""
    fresh = MCPClientManager()
    with (
        patch("docketeer_mcp.tools.manager", fresh),
        patch("docketeer_mcp.prompt.manager", fresh),
    ):
        yield fresh


@pytest.fixture()
def data_dir(tmp_path: Path) -> Generator[Path]:
    d = tmp_path / "data"
    d.mkdir()
    with patch("docketeer_mcp.config.environment") as mock_env:
        mock_env.DATA_DIR = d
        yield d


@pytest.fixture()
def mcp_dir(data_dir: Path) -> Path:
    d = data_dir / "mcp"
    d.mkdir()
    return d


def _write_server(mcp_dir: Path, name: str, data: dict) -> None:
    (mcp_dir / f"{name}.json").write_text(json.dumps(data))


async def test_connect_with_vault_auth(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When config has auth and vault is available, resolve token and connect."""
    _write_server(
        mcp_dir, "api", {"url": "https://api.example.com/mcp", "auth": "mcp/api/token"}
    )

    mock_vault = AsyncMock(spec=Vault)
    mock_vault.resolve = AsyncMock(return_value="bearer_token_123")
    tool_context.vault = mock_vault

    tools = [
        MCPToolInfo(server="api", name="query", description="Query", input_schema={})
    ]
    fresh_manager.connect = AsyncMock(return_value=tools)  # type: ignore[method-assign]

    result = await registry.execute("connect_mcp_server", {"name": "api"}, tool_context)
    assert "1 tools" in result
    fresh_manager.connect.assert_called_once()  # type: ignore[union-attr]
    call_kwargs = fresh_manager.connect.call_args  # type: ignore[union-attr]
    assert call_kwargs.kwargs.get("auth") == "bearer_token_123"


async def test_connect_with_auth_no_vault(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When config has auth but no vault, return error."""
    _write_server(
        mcp_dir, "api", {"url": "https://api.example.com/mcp", "auth": "mcp/api/token"}
    )
    tool_context.vault = None

    result = await registry.execute("connect_mcp_server", {"name": "api"}, tool_context)
    assert "vault" in result.lower()


async def test_connect_http_auth_required(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When HTTP server requires auth, return authorization URL."""
    _write_server(mcp_dir, "github", {"url": "https://mcp.github.com/mcp"})

    with (
        patch("docketeer_mcp.tools._check_auth_required", return_value=True),
        patch(
            "docketeer_mcp.tools.discover_oauth_metadata",
            return_value=(
                "https://auth.github.com/authorize",
                "https://auth.github.com/token",
                "https://auth.github.com/register",
                "read write",
            ),
        ),
        patch(
            "docketeer_mcp.tools.register_client",
            return_value=("client_abc", "secret_xyz"),
        ),
    ):
        result = await registry.execute(
            "connect_mcp_server", {"name": "github"}, tool_context
        )

    assert "Authorization needed" in result
    assert "mcp_oauth_complete" in result
    assert "github" in fresh_manager._pending_oauth


async def test_connect_http_auth_required_no_registration(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When no registration endpoint, use provided client_id."""
    _write_server(mcp_dir, "api", {"url": "https://api.example.com/mcp"})

    with (
        patch("docketeer_mcp.tools._check_auth_required", return_value=True),
        patch(
            "docketeer_mcp.tools.discover_oauth_metadata",
            return_value=(
                "https://api.example.com/authorize",
                "https://api.example.com/token",
                None,
                "read",
            ),
        ),
    ):
        result = await registry.execute(
            "connect_mcp_server",
            {"name": "api", "client_id": "my_client", "client_secret": "my_secret"},
            tool_context,
        )

    assert "Authorization needed" in result
    pending = fresh_manager._pending_oauth["api"]
    assert pending.client_id == "my_client"
    assert pending.client_secret == "my_secret"


async def test_connect_http_no_auth_required(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When HTTP server doesn't require auth, connect normally."""
    _write_server(mcp_dir, "open", {"url": "https://open.example.com/mcp"})
    tools = [
        MCPToolInfo(server="open", name="ping", description="Ping", input_schema={})
    ]
    fresh_manager.connect = AsyncMock(return_value=tools)  # type: ignore[method-assign]

    with patch("docketeer_mcp.tools._check_auth_required", return_value=False):
        result = await registry.execute(
            "connect_mcp_server", {"name": "open"}, tool_context
        )

    assert "1 tools" in result


async def test_mcp_oauth_complete_success(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """Happy path: exchange code, store token, update config."""
    _write_server(mcp_dir, "github", {"url": "https://mcp.github.com/mcp"})

    mock_vault = AsyncMock(spec=Vault)
    tool_context.vault = mock_vault

    fresh_manager._pending_oauth["github"] = PendingOAuth(
        server_url="https://mcp.github.com/mcp",
        authorization_endpoint="https://auth.github.com/authorize",
        token_endpoint="https://auth.github.com/token",
        code_verifier="verifier",
        code_challenge="challenge",
        state="state123",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="client_abc",
    )

    with patch(
        "docketeer_mcp.tools.exchange_code",
        return_value={
            "access_token": "at_new",
            "token_type": "Bearer",
            "expires_in": 3600,
        },
    ):
        result = await registry.execute(
            "mcp_oauth_complete",
            {
                "server": "github",
                "redirect_url": "http://127.0.0.1:3141/callback?code=authcode&state=state123",
                "token_secret": "mcp/github/token",
            },
            tool_context,
        )

    assert (
        "stored" in result.lower()
        or "success" in result.lower()
        or "complete" in result.lower()
    )
    mock_vault.store.assert_any_call("mcp/github/token", "at_new")
    assert "github" not in fresh_manager._pending_oauth


async def test_mcp_oauth_complete_with_refresh(
    tool_context: ToolContext, mcp_dir: Path, fresh_manager: MCPClientManager
):
    """When refresh_token present, store it and schedule refresh task."""
    _write_server(mcp_dir, "api", {"url": "https://api.example.com/mcp"})

    mock_vault = AsyncMock(spec=Vault)
    tool_context.vault = mock_vault

    mock_docket = AsyncMock()
    mock_docket.add = lambda *a, **kw: AsyncMock()

    fresh_manager._pending_oauth["api"] = PendingOAuth(
        server_url="https://api.example.com/mcp",
        authorization_endpoint="https://api.example.com/authorize",
        token_endpoint="https://api.example.com/token",
        code_verifier="verifier",
        code_challenge="challenge",
        state="mystate",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="cid",
        client_secret="csecret",
    )

    with (
        patch(
            "docketeer_mcp.tools.exchange_code",
            return_value={
                "access_token": "at_new",
                "token_type": "Bearer",
                "expires_in": 3600,
                "refresh_token": "rt_new",
            },
        ),
        patch("docketeer_mcp.tools.current_docket", return_value=mock_docket),
    ):
        await registry.execute(
            "mcp_oauth_complete",
            {
                "server": "api",
                "redirect_url": "http://127.0.0.1:3141/callback?code=code1&state=mystate",
                "token_secret": "mcp/api/token",
            },
            tool_context,
        )

    mock_vault.store.assert_any_call("mcp/api/token", "at_new")
    mock_vault.store.assert_any_call("mcp/api/token/refresh", "rt_new")
    mock_vault.store.assert_any_call("mcp/api/token/client_id", "cid")
    mock_vault.store.assert_any_call("mcp/api/token/client_secret", "csecret")
    mock_vault.store.assert_any_call(
        "mcp/api/token/token_endpoint", "https://api.example.com/token"
    )


async def test_mcp_oauth_complete_state_mismatch(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    """State mismatch returns error."""
    mock_vault = AsyncMock(spec=Vault)
    tool_context.vault = mock_vault

    fresh_manager._pending_oauth["s"] = PendingOAuth(
        server_url="https://example.com/mcp",
        authorization_endpoint="https://example.com/authorize",
        token_endpoint="https://example.com/token",
        code_verifier="v",
        code_challenge="c",
        state="expected_state",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="cid",
    )

    result = await registry.execute(
        "mcp_oauth_complete",
        {
            "server": "s",
            "redirect_url": "http://127.0.0.1:3141/callback?code=x&state=wrong_state",
            "token_secret": "mcp/s/token",
        },
        tool_context,
    )
    assert "state" in result.lower()


async def test_mcp_oauth_complete_no_pending(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    """No pending OAuth for server returns error."""
    mock_vault = AsyncMock(spec=Vault)
    tool_context.vault = mock_vault

    result = await registry.execute(
        "mcp_oauth_complete",
        {
            "server": "unknown",
            "redirect_url": "http://127.0.0.1:3141/callback?code=x&state=y",
            "token_secret": "mcp/x/token",
        },
        tool_context,
    )
    assert "no pending" in result.lower() or "not found" in result.lower()


async def test_mcp_oauth_complete_no_vault(
    tool_context: ToolContext, fresh_manager: MCPClientManager
):
    """No vault returns error."""
    tool_context.vault = None

    result = await registry.execute(
        "mcp_oauth_complete",
        {
            "server": "s",
            "redirect_url": "http://127.0.0.1:3141/callback?code=x&state=y",
            "token_secret": "mcp/s/token",
        },
        tool_context,
    )
    assert "vault" in result.lower()


async def test_check_auth_required_401():
    """Returns True when server responds with 401."""
    from docketeer_mcp.tools import _check_auth_required

    mock_response = httpx.Response(
        status_code=401,
        request=httpx.Request("POST", "https://example.com/mcp"),
    )

    async def mock_post(url: str, **kwargs: object) -> httpx.Response:
        return mock_response

    mock_client = AsyncMock()
    mock_client.post = mock_post
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.tools.httpx.AsyncClient", return_value=mock_client):
        assert await _check_auth_required("https://example.com/mcp") is True


async def test_check_auth_required_200():
    """Returns False when server responds with 200."""
    from docketeer_mcp.tools import _check_auth_required

    mock_response = httpx.Response(
        status_code=200,
        request=httpx.Request("POST", "https://example.com/mcp"),
    )

    async def mock_post(url: str, **kwargs: object) -> httpx.Response:
        return mock_response

    mock_client = AsyncMock()
    mock_client.post = mock_post
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.tools.httpx.AsyncClient", return_value=mock_client):
        assert await _check_auth_required("https://example.com/mcp") is False


async def test_check_auth_required_exception():
    """Returns False when request raises an exception."""
    from docketeer_mcp.tools import _check_auth_required

    mock_client = AsyncMock()
    mock_client.post = AsyncMock(side_effect=httpx.ConnectError("refused"))
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.tools.httpx.AsyncClient", return_value=mock_client):
        assert await _check_auth_required("https://example.com/mcp") is False


def test_current_docket():
    """Resolves docket from context var."""
    from contextvars import copy_context

    from docketeer.dependencies import set_docket
    from docketeer_mcp.tools import current_docket

    mock_docket = AsyncMock()

    def _run() -> Docket:
        set_docket(mock_docket)
        return current_docket()

    ctx = copy_context()
    result = ctx.run(_run)
    assert result is mock_docket



"""Tests for MCP OAuth helpers."""

import json
from unittest.mock import AsyncMock, patch

import httpx
import pytest

from docketeer_mcp.oauth import (
    PendingOAuth,
    build_authorization_url,
    discover_oauth_metadata,
    exchange_code,
    refresh_access_token,
    register_client,
)


def _mock_response(
    status_code: int = 200, json_data: dict | None = None
) -> httpx.Response:
    """Build a fake httpx.Response."""
    content = json.dumps(json_data or {}).encode()
    return httpx.Response(
        status_code=status_code,
        content=content,
        headers={"content-type": "application/json"},
        request=httpx.Request("GET", "https://example.com"),
    )


async def test_discover_with_prm_and_oasm():
    prm_response = _mock_response(
        json_data={
            "resource": "https://mcp.example.com/mcp",
            "authorization_servers": ["https://auth.example.com"],
        }
    )
    oasm_response = _mock_response(
        json_data={
            "issuer": "https://auth.example.com",
            "authorization_endpoint": "https://auth.example.com/authorize",
            "token_endpoint": "https://auth.example.com/token",
            "registration_endpoint": "https://auth.example.com/register",
            "scopes_supported": ["read", "write"],
        }
    )

    responses = iter([prm_response, oasm_response])

    async def mock_get(url: str, **kwargs: object) -> httpx.Response:
        return next(responses)

    mock_client = AsyncMock()
    mock_client.get = mock_get
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        auth_ep, token_ep, reg_ep, scopes = await discover_oauth_metadata(
            "https://mcp.example.com/mcp"
        )

    assert auth_ep == "https://auth.example.com/authorize"
    assert token_ep == "https://auth.example.com/token"
    assert reg_ep == "https://auth.example.com/register"
    assert scopes == "read write"


async def test_discover_prm_no_scopes():
    """PRM without scopes_supported still works, scopes come from OASM."""
    prm_response = _mock_response(
        json_data={
            "resource": "https://mcp.example.com/mcp",
            "authorization_servers": ["https://auth.example.com"],
        }
    )
    oasm_response = _mock_response(
        json_data={
            "issuer": "https://auth.example.com",
            "authorization_endpoint": "https://auth.example.com/authorize",
            "token_endpoint": "https://auth.example.com/token",
            "scopes_supported": ["api"],
        }
    )

    responses = iter([prm_response, oasm_response])

    async def mock_get(url: str, **kwargs: object) -> httpx.Response:
        return next(responses)

    mock_client = AsyncMock()
    mock_client.get = mock_get
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        auth_ep, token_ep, reg_ep, scopes = await discover_oauth_metadata(
            "https://mcp.example.com/mcp"
        )

    assert scopes == "api"


async def test_discover_with_pathed_auth_server():
    """Auth server URL with a path uses path-aware OASM discovery."""
    prm_response = _mock_response(
        json_data={
            "resource": "https://mcp.example.com/mcp",
            "authorization_servers": ["https://auth.example.com/tenant/abc"],
        }
    )
    oasm_response = _mock_response(
        json_data={
            "issuer": "https://auth.example.com/tenant/abc",
            "authorization_endpoint": "https://auth.example.com/tenant/abc/authorize",
            "token_endpoint": "https://auth.example.com/tenant/abc/token",
        }
    )

    responses = iter([prm_response, oasm_response])

    async def mock_get(url: str, **kwargs: object) -> httpx.Response:
        return next(responses)

    mock_client = AsyncMock()
    mock_client.get = mock_get
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        auth_ep, token_ep, reg_ep, scopes = await discover_oauth_metadata(
            "https://mcp.example.com/mcp"
        )

    assert auth_ep == "https://auth.example.com/tenant/abc/authorize"
    assert token_ep == "https://auth.example.com/tenant/abc/token"


async def test_discover_root_url_no_path():
    """Server at root URL (no path) only tries root PRM."""
    prm_response = _mock_response(
        json_data={
            "resource": "https://mcp.example.com/",
            "authorization_servers": ["https://auth.example.com"],
        }
    )
    oasm_response = _mock_response(
        json_data={
            "issuer": "https://auth.example.com",
            "authorization_endpoint": "https://auth.example.com/authorize",
            "token_endpoint": "https://auth.example.com/token",
        }
    )

    responses = iter([prm_response, oasm_response])

    async def mock_get(url: str, **kwargs: object) -> httpx.Response:
        return next(responses)

    mock_client = AsyncMock()
    mock_client.get = mock_get
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        auth_ep, token_ep, reg_ep, scopes = await discover_oauth_metadata(
            "https://mcp.example.com/"
        )

    assert auth_ep == "https://auth.example.com/authorize"


async def test_discover_legacy_no_prm():
    """When PRM returns 404, fall back to legacy OASM discovery."""
    prm_404_a = _mock_response(status_code=404)
    prm_404_b = _mock_response(status_code=404)
    oasm_response = _mock_response(
        json_data={
            "issuer": "https://mcp.example.com",
            "authorization_endpoint": "https://mcp.example.com/authorize",
            "token_endpoint": "https://mcp.example.com/token",
        }
    )

    responses = iter([prm_404_a, prm_404_b, oasm_response])

    async def mock_get(url: str, **kwargs: object) -> httpx.Response:
        return next(responses)

    mock_client = AsyncMock()
    mock_client.get = mock_get
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        auth_ep, token_ep, reg_ep, scopes = await discover_oauth_metadata(
            "https://mcp.example.com/mcp"
        )

    assert auth_ep == "https://mcp.example.com/authorize"
    assert token_ep == "https://mcp.example.com/token"
    assert reg_ep is None
    assert scopes is None


async def test_discover_prm_with_scopes():
    """PRM includes scopes_supported — scopes come from PRM."""
    prm_response = _mock_response(
        json_data={
            "resource": "https://mcp.example.com/mcp",
            "authorization_servers": ["https://auth.example.com"],
            "scopes_supported": ["read", "write", "admin"],
        }
    )
    oasm_response = _mock_response(
        json_data={
            "issuer": "https://auth.example.com",
            "authorization_endpoint": "https://auth.example.com/authorize",
            "token_endpoint": "https://auth.example.com/token",
        }
    )

    responses = iter([prm_response, oasm_response])

    async def mock_get(url: str, **kwargs: object) -> httpx.Response:
        return next(responses)

    mock_client = AsyncMock()
    mock_client.get = mock_get
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        auth_ep, token_ep, reg_ep, scopes = await discover_oauth_metadata(
            "https://mcp.example.com/mcp"
        )

    assert scopes == "read write admin"
    assert auth_ep == "https://auth.example.com/authorize"


async def test_discover_oasm_fails():
    """When both PRM and OASM return 404, raise."""
    prm_404_a = _mock_response(status_code=404)
    prm_404_b = _mock_response(status_code=404)
    oasm_404 = _mock_response(status_code=404)

    responses = iter([prm_404_a, prm_404_b, oasm_404])

    async def mock_get(url: str, **kwargs: object) -> httpx.Response:
        return next(responses)

    mock_client = AsyncMock()
    mock_client.get = mock_get
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        with pytest.raises(RuntimeError, match="Could not discover"):
            await discover_oauth_metadata("https://mcp.example.com/mcp")


async def test_register_client():
    response = _mock_response(
        json_data={
            "client_id": "abc123",
            "client_secret": "secret456",
        }
    )

    async def mock_post(url: str, **kwargs: object) -> httpx.Response:
        return response

    mock_client = AsyncMock()
    mock_client.post = mock_post
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        client_id, client_secret = await register_client(
            "https://auth.example.com/register",
            "http://127.0.0.1:3141/callback",
            "docketeer",
            "read write",
        )

    assert client_id == "abc123"
    assert client_secret == "secret456"


async def test_register_client_no_secret():
    response = _mock_response(
        json_data={
            "client_id": "abc123",
        }
    )

    async def mock_post(url: str, **kwargs: object) -> httpx.Response:
        return response

    mock_client = AsyncMock()
    mock_client.post = mock_post
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        client_id, client_secret = await register_client(
            "https://auth.example.com/register",
            "http://127.0.0.1:3141/callback",
            "docketeer",
        )

    assert client_id == "abc123"
    assert client_secret == ""


async def test_register_client_failure():
    response = _mock_response(
        status_code=400, json_data={"error": "invalid_client_metadata"}
    )

    async def mock_post(url: str, **kwargs: object) -> httpx.Response:
        return response

    mock_client = AsyncMock()
    mock_client.post = mock_post
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        with pytest.raises(RuntimeError, match="Client registration failed"):
            await register_client(
                "https://auth.example.com/register",
                "http://127.0.0.1:3141/callback",
                "docketeer",
            )


def test_build_authorization_url():
    pending = PendingOAuth(
        server_url="https://mcp.example.com/mcp",
        authorization_endpoint="https://auth.example.com/authorize",
        token_endpoint="https://auth.example.com/token",
        code_verifier="test_verifier_that_is_long_enough_for_pkce_requirements_43chars",
        code_challenge="test_challenge",
        state="random_state",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="abc123",
        scopes="read write",
    )
    url = build_authorization_url(pending)

    assert "https://auth.example.com/authorize?" in url
    assert "response_type=code" in url
    assert "client_id=abc123" in url
    assert "state=random_state" in url
    assert "code_challenge=test_challenge" in url
    assert "code_challenge_method=S256" in url
    assert "scope=read+write" in url
    assert "redirect_uri=" in url


def test_build_authorization_url_no_scopes():
    pending = PendingOAuth(
        server_url="https://mcp.example.com/mcp",
        authorization_endpoint="https://auth.example.com/authorize",
        token_endpoint="https://auth.example.com/token",
        code_verifier="test_verifier_that_is_long_enough_for_pkce_requirements_43chars",
        code_challenge="test_challenge",
        state="random_state",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="abc123",
    )
    url = build_authorization_url(pending)

    assert "scope=" not in url


async def test_exchange_code():
    response = _mock_response(
        json_data={
            "access_token": "at_123",
            "token_type": "Bearer",
            "expires_in": 3600,
            "refresh_token": "rt_456",
        }
    )

    async def mock_post(url: str, **kwargs: object) -> httpx.Response:
        return response

    mock_client = AsyncMock()
    mock_client.post = mock_post
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    pending = PendingOAuth(
        server_url="https://mcp.example.com/mcp",
        authorization_endpoint="https://auth.example.com/authorize",
        token_endpoint="https://auth.example.com/token",
        code_verifier="verifier_long_enough_for_pkce_requirements_here_43chars",
        code_challenge="challenge",
        state="state123",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="abc123",
        client_secret="secret",
    )

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        tokens = await exchange_code(pending, "auth_code_xyz")

    assert tokens["access_token"] == "at_123"
    assert tokens["refresh_token"] == "rt_456"
    assert tokens["expires_in"] == 3600


async def test_exchange_code_failure():
    response = _mock_response(status_code=400, json_data={"error": "invalid_grant"})

    async def mock_post(url: str, **kwargs: object) -> httpx.Response:
        return response

    mock_client = AsyncMock()
    mock_client.post = mock_post
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    pending = PendingOAuth(
        server_url="https://mcp.example.com/mcp",
        authorization_endpoint="https://auth.example.com/authorize",
        token_endpoint="https://auth.example.com/token",
        code_verifier="verifier_long_enough_for_pkce_requirements_here_43chars",
        code_challenge="challenge",
        state="state123",
        redirect_uri="http://127.0.0.1:3141/callback",
        client_id="abc123",
    )

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        with pytest.raises(RuntimeError, match="Token exchange failed"):
            await exchange_code(pending, "bad_code")


async def test_refresh_access_token():
    response = _mock_response(
        json_data={
            "access_token": "new_at",
            "token_type": "Bearer",
            "expires_in": 7200,
            "refresh_token": "new_rt",
        }
    )

    async def mock_post(url: str, **kwargs: object) -> httpx.Response:
        return response

    mock_client = AsyncMock()
    mock_client.post = mock_post
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        tokens = await refresh_access_token(
            "https://auth.example.com/token",
            "old_refresh_token",
            "abc123",
            "secret",
        )

    assert tokens["access_token"] == "new_at"
    assert tokens["refresh_token"] == "new_rt"


async def test_refresh_access_token_failure():
    response = _mock_response(status_code=400, json_data={"error": "invalid_grant"})

    async def mock_post(url: str, **kwargs: object) -> httpx.Response:
        return response

    mock_client = AsyncMock()
    mock_client.post = mock_post
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=None)

    with patch("docketeer_mcp.oauth.httpx.AsyncClient", return_value=mock_client):
        with pytest.raises(RuntimeError, match="Token refresh failed"):
            await refresh_access_token(
                "https://auth.example.com/token",
                "bad_refresh",
                "abc123",
            )



"""Tests for the MCP prompt provider."""

import json
from collections.abc import Generator
from pathlib import Path
from unittest.mock import patch

import pytest

from docketeer_mcp.manager import MCPClientManager, MCPToolInfo
from docketeer_mcp.prompt import provide_mcp_catalog


@pytest.fixture(autouse=True)
def fresh_manager() -> Generator[MCPClientManager]:
    fresh = MCPClientManager()
    with patch("docketeer_mcp.prompt.manager", fresh):
        yield fresh


@pytest.fixture()
def data_dir(tmp_path: Path) -> Generator[Path]:
    d = tmp_path / "data"
    d.mkdir()
    with patch("docketeer_mcp.config.environment") as mock_env:
        mock_env.DATA_DIR = d
        yield d


@pytest.fixture()
def mcp_dir(data_dir: Path) -> Path:
    d = data_dir / "mcp"
    d.mkdir()
    return d


def test_no_servers(workspace: Path, data_dir: Path):
    assert provide_mcp_catalog(workspace) == []


def test_with_servers_disconnected(workspace: Path, mcp_dir: Path):
    (mcp_dir / "time.json").write_text(
        json.dumps({"command": "uvx", "args": ["mcp-server-time"]})
    )
    (mcp_dir / "api.json").write_text(
        json.dumps({"url": "https://api.example.com/mcp"})
    )

    blocks = provide_mcp_catalog(workspace)
    assert len(blocks) == 1
    text = blocks[0].text
    assert "## MCP Servers" in text
    assert "**time**: `uvx`" in text
    assert "**api**: `https://api.example.com/mcp`" in text
    assert "connect_mcp_server" in text
    assert "Connected:" not in text


def test_with_connected_server(
    workspace: Path, mcp_dir: Path, fresh_manager: MCPClientManager
):
    (mcp_dir / "time.json").write_text(json.dumps({"command": "uvx"}))
    fresh_manager._tools["time"] = [
        MCPToolInfo(server="time", name="t", description="", input_schema={}),
        MCPToolInfo(server="time", name="u", description="", input_schema={}),
    ]
    fresh_manager._clients["time"] = object()  # type: ignore[assignment]

    blocks = provide_mcp_catalog(workspace)
    text = blocks[0].text
    assert "Connected: time (2 tools)" in text



"""MCP client manager — connections, tool catalog, and dispatch."""

import logging
from dataclasses import dataclass
from pathlib import Path

from fastmcp import Client
from fastmcp.client.transports import ClientTransport, StreamableHttpTransport
from fastmcp.client.transports.stdio import StdioTransport

from docketeer.executor import CommandExecutor, Mount

from .config import MCPServerConfig
from .oauth import PendingOAuth
from .transport import ExecutorTransport

log = logging.getLogger(__name__)


@dataclass
class MCPToolInfo:
    """Metadata for a single tool on a connected MCP server."""

    server: str
    name: str
    description: str
    input_schema: dict


def _build_transport(
    config: MCPServerConfig,
    executor: CommandExecutor | None = None,
    workspace: Path | None = None,
    auth: str | None = None,
) -> ClientTransport:
    """Build the appropriate transport for a server config."""
    if config.is_stdio:
        command = [config.command, *config.args]
        if executor:
            mounts: list[Mount] = []
            if workspace:
                mounts.append(Mount(source=workspace, target=Path("/workspace")))
            return ExecutorTransport(
                executor=executor,
                command=command,
                env=config.env or None,
                mounts=mounts or None,
                network_access=config.network_access,
            )
        log.warning(
            "No executor available — running MCP server %r without sandbox",
            config.name,
        )
        return StdioTransport(
            command=config.command,
            args=config.args,
            env=config.env or None,
        )

    if config.is_http:
        return StreamableHttpTransport(
            url=config.url,
            headers=config.headers or None,
            auth=auth,
        )

    raise ValueError(f"Server {config.name!r} has neither command nor url")


class MCPClientManager:
    """Manages connections to MCP servers and their tool catalogs."""

    def __init__(self) -> None:
        self._clients: dict[str, Client] = {}
        self._tools: dict[str, list[MCPToolInfo]] = {}
        self._pending_oauth: dict[str, PendingOAuth] = {}

    async def connect(
        self,
        name: str,
        config: MCPServerConfig,
        executor: CommandExecutor | None = None,
        workspace: Path | None = None,
        auth: str | None = None,
    ) -> list[MCPToolInfo]:
        """Connect to a server and discover its tools."""
        if name in self._clients:
            raise ValueError(f"Already connected to {name!r}")

        transport = _build_transport(config, executor, workspace, auth=auth)
        client = Client(transport)
        await client.__aenter__()

        try:
            raw_tools = await client.list_tools()
        except Exception:
            await client.__aexit__(None, None, None)
            raise

        tools = [
            MCPToolInfo(
                server=name,
                name=t.name,
                description=t.description or "",
                input_schema=t.inputSchema,
            )
            for t in raw_tools
        ]
        self._clients[name] = client
        self._tools[name] = tools
        return tools

    async def disconnect(self, name: str) -> None:
        """Disconnect from a server."""
        client = self._clients.pop(name, None)
        self._tools.pop(name, None)
        if client:
            await client.__aexit__(None, None, None)

    async def disconnect_all(self) -> None:
        """Disconnect from all servers."""
        for name in list(self._clients):
            await self.disconnect(name)

    def is_connected(self, name: str) -> bool:
        return name in self._clients

    def connected_servers(self) -> dict[str, int]:
        """Return connected server names with their tool counts."""
        return {name: len(tools) for name, tools in self._tools.items()}

    def search_tools(self, query: str, server: str = "") -> list[MCPToolInfo]:
        """Search tools by substring match on name or description."""
        query_lower = query.lower()
        results: list[MCPToolInfo] = []
        sources = (
            {server: self._tools[server]}
            if server and server in self._tools
            else self._tools
        )
        for tools in sources.values():
            for tool in tools:
                if (
                    query_lower in tool.name.lower()
                    or query_lower in tool.description.lower()
                ):
                    results.append(tool)
        return results

    async def call_tool(self, server: str, tool: str, arguments: dict) -> str:
        """Call a tool on a connected server, returning the result as a string."""
        client = self._clients.get(server)
        if not client:
            raise ValueError(f"Not connected to server {server!r}")

        result = await client.call_tool(tool, arguments)
        parts = []
        for block in result.content:
            if hasattr(block, "text"):
                parts.append(block.text)
            else:
                parts.append(str(block))
        return "\n".join(parts)


manager = MCPClientManager()



"""Docket task functions for MCP OAuth token refresh."""

import logging
from datetime import datetime, timedelta

from docket import Docket

from docketeer.dependencies import CurrentDocket, CurrentVault
from docketeer.vault import Vault

from .oauth import refresh_access_token

log = logging.getLogger(__name__)


async def mcp_oauth_refresh(
    token_secret: str,
    token_endpoint: str,
    client_id: str,
    client_secret: str = "",
    expires_in: int = 3600,
    vault: Vault = CurrentVault(),
    docket: Docket = CurrentDocket(),
) -> None:
    """Refresh an OAuth access token and reschedule for the next refresh."""
    try:
        refresh_token = await vault.resolve(f"{token_secret}/refresh")
    except Exception:
        log.exception("Could not resolve refresh token for %s", token_secret)
        return

    try:
        tokens = await refresh_access_token(
            token_endpoint, refresh_token, client_id, client_secret
        )
    except Exception:
        log.exception("Token refresh failed for %s", token_secret)
        return

    access_token = str(tokens.get("access_token", ""))
    if access_token:
        await vault.store(token_secret, access_token)

    new_refresh = str(tokens.get("refresh_token", ""))
    if new_refresh:
        await vault.store(f"{token_secret}/refresh", new_refresh)

    new_expires = int(str(tokens.get("expires_in", expires_in)))

    fire_at = datetime.now().astimezone() + timedelta(
        seconds=max(new_expires - 300, 60)
    )
    await docket.add(
        mcp_oauth_refresh, when=fire_at, key=f"mcp-refresh-{token_secret}"
    )(
        token_secret=token_secret,
        token_endpoint=token_endpoint,
        client_id=client_id,
        client_secret=client_secret,
        expires_in=new_expires,
    )


mcp_tasks = [mcp_oauth_refresh]

mcp_task_collections = ["docketeer_mcp.tasks:mcp_tasks"]



"""Agent-facing MCP tools."""

import json
import logging
import secrets
from urllib.parse import parse_qs, urlparse

import httpx
from docket import Docket

from docketeer.tools import ToolContext, registry

from . import config
from .manager import manager
from .oauth import (
    REDIRECT_URI,
    PendingOAuth,
    _generate_pkce,
    build_authorization_url,
    discover_oauth_metadata,
    exchange_code,
    register_client,
)

log = logging.getLogger(__name__)


async def _check_auth_required(url: str) -> bool:
    """Probe an HTTP MCP server to see if it returns 401."""
    try:
        async with httpx.AsyncClient() as client:
            resp = await client.post(url, timeout=10)
            return resp.status_code == 401
    except Exception:
        return False


def current_docket() -> Docket:
    """Resolve the current Docket from dependencies."""
    from docketeer.dependencies import _docket_var

    return _docket_var.get()


@registry.tool
async def list_mcp_servers(ctx: ToolContext) -> str:
    """List configured MCP servers and their connection status."""
    servers = config.load_servers()
    if not servers:
        return "No MCP servers configured."

    connected = manager.connected_servers()
    lines = []
    for name, cfg in servers.items():
        kind = cfg.command if cfg.is_stdio else cfg.url
        status = (
            f"connected ({connected[name]} tools)"
            if name in connected
            else "disconnected"
        )
        lines.append(f"- **{name}**: `{kind}` — {status}")
    return "\n".join(lines)


@registry.tool
async def connect_mcp_server(
    ctx: ToolContext,
    name: str,
    client_id: str = "",
    client_secret: str = "",
    scopes: str = "",
) -> str:
    """Connect to a configured MCP server and discover its tools. If the
    server requires OAuth, returns an authorization URL for the user.

    name: server name from the configuration
    client_id: pre-configured OAuth client ID (optional, skips registration)
    client_secret: pre-configured OAuth client secret (optional)
    scopes: OAuth scopes to request (optional)
    """
    if manager.is_connected(name):
        return f"Already connected to {name!r}."

    servers = config.load_servers()
    cfg = servers.get(name)
    if not cfg:
        return f"No server configured with name {name!r}."

    # If config has an auth secret, resolve from vault and connect with bearer
    if cfg.auth:
        if not ctx.vault:
            return f"Server {name!r} requires auth but no vault is configured."
        try:
            token = await ctx.vault.resolve(cfg.auth)
        except Exception as e:
            return f"Failed to resolve auth secret {cfg.auth!r}: {e}"

        try:
            tools = await manager.connect(
                name, cfg, ctx.executor, ctx.workspace, auth=token
            )
        except Exception as e:
            log.warning("Failed to connect to MCP server %r", name, exc_info=True)
            return f"Failed to connect to {name!r}: {e}"

        if not tools:
            return f"Connected to {name!r} — no tools found."
        lines = [f"Connected to {name!r} — {len(tools)} tools:"]
        for t in tools:
            lines.append(f"- **{t.name}**: {t.description}")
        return "\n".join(lines)

    # For HTTP servers without auth config, check if auth is required
    if cfg.is_http:
        auth_required = await _check_auth_required(cfg.url)
        if auth_required:
            return await _start_oauth_flow(name, cfg, client_id, client_secret, scopes)

    # No auth needed — connect normally
    try:
        tools = await manager.connect(name, cfg, ctx.executor, ctx.workspace)
    except Exception as e:
        log.warning("Failed to connect to MCP server %r", name, exc_info=True)
        return f"Failed to connect to {name!r}: {e}"

    if not tools:
        return f"Connected to {name!r} — no tools found."

    lines = [f"Connected to {name!r} — {len(tools)} tools:"]
    for t in tools:
        lines.append(f"- **{t.name}**: {t.description}")
    return "\n".join(lines)


async def _start_oauth_flow(
    name: str,
    cfg: config.MCPServerConfig,
    client_id: str,
    client_secret: str,
    scopes: str,
) -> str:
    """Discover OAuth metadata, register if needed, store pending state."""
    try:
        auth_ep, token_ep, reg_ep, discovered_scopes = await discover_oauth_metadata(
            cfg.url
        )
    except Exception as e:
        return f"OAuth discovery failed for {name!r}: {e}"

    effective_scopes = scopes or discovered_scopes or ""

    if not client_id:
        if reg_ep:
            try:
                client_id, client_secret = await register_client(
                    reg_ep, REDIRECT_URI, f"docketeer-{name}", effective_scopes
                )
            except Exception as e:
                return f"Client registration failed for {name!r}: {e}"
        else:
            return (
                f"Server {name!r} requires OAuth but has no registration endpoint. "
                f"Provide client_id (and client_secret) when connecting."
            )

    verifier, challenge = _generate_pkce()
    state = secrets.token_urlsafe(32)

    pending = PendingOAuth(
        server_url=cfg.url,
        authorization_endpoint=auth_ep,
        token_endpoint=token_ep,
        code_verifier=verifier,
        code_challenge=challenge,
        state=state,
        redirect_uri=REDIRECT_URI,
        client_id=client_id,
        client_secret=client_secret,
        scopes=effective_scopes,
    )

    manager._pending_oauth[name] = pending
    auth_url = build_authorization_url(pending)

    return (
        f"Authorization needed for {name!r}.\n\n"
        f"Send this URL to the user:\n{auth_url}\n\n"
        f"After they authorize, use `mcp_oauth_complete` with the redirect URL."
    )


@registry.tool
async def mcp_oauth_complete(
    ctx: ToolContext,
    server: str,
    redirect_url: str,
    token_secret: str,
) -> str:
    """Complete an OAuth flow by exchanging the authorization code for tokens.

    server: the MCP server name from the pending OAuth flow
    redirect_url: the full redirect URL the user was sent to (contains code and state)
    token_secret: vault secret path to store the access token (e.g. mcp/github/token)
    """
    if not ctx.vault:
        return "No vault configured — cannot store OAuth tokens."

    pending = manager._pending_oauth.get(server)
    if not pending:
        return f"No pending OAuth flow found for {server!r}."

    # Parse code and state from redirect URL
    parsed = urlparse(redirect_url)
    params = parse_qs(parsed.query)
    code = params.get("code", [""])[0]
    state = params.get("state", [""])[0]

    if not code:
        return "No authorization code found in the redirect URL."

    if not secrets.compare_digest(state, pending.state):
        return "State mismatch — the redirect URL doesn't match the pending flow."

    # Exchange code for tokens
    try:
        tokens = await exchange_code(pending, code)
    except Exception as e:
        return f"Token exchange failed: {e}"

    access_token = str(tokens.get("access_token", ""))
    if not access_token:
        return "Token exchange succeeded but no access_token in response."

    # Store access token in vault
    await ctx.vault.store(token_secret, access_token)

    # Update server config with auth secret path
    servers = config.load_servers()
    cfg = servers.get(server)
    if cfg:
        cfg.auth = token_secret
        config.save_server(cfg)

    # Clean up pending state
    del manager._pending_oauth[server]

    # Handle refresh token if present
    refresh_token = str(tokens.get("refresh_token", ""))
    if refresh_token:
        await ctx.vault.store(f"{token_secret}/refresh", refresh_token)
        await ctx.vault.store(f"{token_secret}/client_id", pending.client_id)
        await ctx.vault.store(f"{token_secret}/client_secret", pending.client_secret)
        await ctx.vault.store(f"{token_secret}/token_endpoint", pending.token_endpoint)

        expires_in = int(str(tokens.get("expires_in", 3600)))
        try:
            await _schedule_token_refresh(
                token_secret,
                pending.token_endpoint,
                pending.client_id,
                pending.client_secret,
                expires_in,
            )
        except Exception:
            log.warning("Could not schedule token refresh", exc_info=True)

    return f"OAuth complete for {server!r}. Token stored at {token_secret!r}."


async def _schedule_token_refresh(
    token_secret: str,
    token_endpoint: str,
    client_id: str,
    client_secret: str,
    expires_in: int,
) -> None:
    """Schedule a one-shot refresh task before the token expires."""
    from datetime import datetime, timedelta

    from .tasks import mcp_oauth_refresh

    docket = current_docket()
    fire_at = datetime.now().astimezone() + timedelta(seconds=max(expires_in - 300, 60))
    await docket.add(
        mcp_oauth_refresh, when=fire_at, key=f"mcp-refresh-{token_secret}"
    )(
        token_secret=token_secret,
        token_endpoint=token_endpoint,
        client_id=client_id,
        client_secret=client_secret,
        expires_in=expires_in,
    )


@registry.tool
async def disconnect_mcp_server(ctx: ToolContext, name: str) -> str:
    """Disconnect from a connected MCP server.

    name: server name to disconnect
    """
    if not manager.is_connected(name):
        return f"Not connected to {name!r}."
    await manager.disconnect(name)
    return f"Disconnected from {name!r}."


@registry.tool
async def search_mcp_tools(ctx: ToolContext, query: str, server: str = "") -> str:
    """Search connected MCP servers for tools matching a query.

    query: search term to match against tool names and descriptions
    server: optional server name to limit the search to
    """
    results = manager.search_tools(query, server=server)
    if not results:
        return f"No tools matching {query!r}."

    lines = []
    for t in results:
        lines.append(f"### {t.server} / {t.name}")
        if t.description:
            lines.append(t.description)
        lines.append(f"```json\n{json.dumps(t.input_schema, indent=2)}\n```")
        lines.append("")
    return "\n".join(lines)


@registry.tool
async def use_mcp_tool(
    ctx: ToolContext, server: str, tool: str, arguments: str = "{}"
) -> str:
    """Call a tool on a connected MCP server.

    server: server name
    tool: tool name on that server
    arguments: JSON string of tool arguments
    """
    try:
        args = json.loads(arguments)
    except json.JSONDecodeError as e:
        return f"Invalid JSON arguments: {e}"

    try:
        return await manager.call_tool(server, tool, args)
    except Exception as e:
        return f"Error calling {server}/{tool}: {e}"


@registry.tool
async def add_mcp_server(
    ctx: ToolContext,
    name: str,
    command: str = "",
    args: str = "[]",
    env: str = "{}",
    url: str = "",
    headers: str = "{}",
    network_access: bool = False,
) -> str:
    """Save a new MCP server configuration.

    name: identifier for the server
    command: executable to run (for stdio servers)
    args: JSON array of command arguments
    env: JSON object of environment variables
    url: server URL (for HTTP servers)
    headers: JSON object of HTTP headers
    network_access: whether the server needs network access (stdio only)
    """
    try:
        args_list = json.loads(args)
    except json.JSONDecodeError as e:
        return f"Invalid args JSON: {e}"

    try:
        env_dict = json.loads(env)
    except json.JSONDecodeError as e:
        return f"Invalid env JSON: {e}"

    try:
        headers_dict = json.loads(headers)
    except json.JSONDecodeError as e:
        return f"Invalid headers JSON: {e}"

    if not command and not url:
        return "Must provide either command (stdio) or url (HTTP)."

    cfg = config.MCPServerConfig(
        name=name,
        command=command,
        args=args_list,
        env=env_dict,
        url=url,
        headers=headers_dict,
        network_access=network_access,
    )
    try:
        config.save_server(cfg)
    except ValueError as e:
        return str(e)

    kind = f"command `{command}`" if command else f"url `{url}`"
    return f"Saved server {name!r} ({kind})."


@registry.tool
async def remove_mcp_server(ctx: ToolContext, name: str) -> str:
    """Remove an MCP server configuration.

    name: server name to remove
    """
    if manager.is_connected(name):
        await manager.disconnect(name)

    servers = config.load_servers()
    cfg = servers.get(name)
    if cfg and cfg.auth:
        try:
            docket = current_docket()
            await docket.cancel(f"mcp-refresh-{cfg.auth}")
        except Exception:
            log.debug("Could not cancel refresh task for %s", name, exc_info=True)

    if config.remove_server(name):
        return f"Removed server {name!r}."
    return f"No server configured with name {name!r}."
